import copy
from unittest.mock import MagicMock

import numpy as np
import pytest
import torch
import torch.nn as nn
import torch.optim as optim
from accelerate import Accelerator
from accelerate.optimizer import AcceleratedOptimizer
from gymnasium import spaces
from gymnasium.spaces import Box, Discrete
from pettingzoo import ParallelEnv
from torch._dynamo import OptimizedModule

from agilerl.algorithms.ippo import IPPO
from agilerl.modules import EvolvableMLP, ModuleDict
from agilerl.modules.custom_components import GumbelSoftmax
from agilerl.networks.actors import StochasticActor
from agilerl.networks.value_networks import ValueNetwork
from agilerl.utils.evolvable_networks import get_default_encoder_config
from agilerl.utils.utils import make_multi_agent_vect_envs
from agilerl.wrappers.make_evolvable import MakeEvolvable
from tests.helper_functions import (
    assert_not_equal_state_dict,
    assert_state_dicts_equal,
    get_sample_from_space,
)


class DummyMultiEnv(ParallelEnv):
    def __init__(self, observation_spaces, action_spaces):
        self.observation_spaces = observation_spaces
        self.action_spaces = action_spaces
        self.agents = ["agent_0", "agent_1", "other_agent_0"]
        self.possible_agents = ["agent_0", "agent_1", "other_agent_0"]
        self.metadata = None
        self.render_mode = None

    def action_space(self, agent):
        return Discrete(self.action_spaces[0].n)

    def observation_space(self, agent):
        return Box(0, 1, self.observation_spaces[0].shape)

    def reset(self, seed=None, options=None):
        return {
            agent: np.random.rand(*self.observation_spaces[i].shape)
            for i, agent in enumerate(self.agents)
        }, {
            "agent_0": {"env_defined_actions": np.array([1])},
            "agent_1": {"env_defined_actions": None},
            "other_agent_0": {"env_defined_actions": None},
        }

    def step(self, action):
        return (
            {
                agent: np.random.rand(*self.observation_spaces[i].shape)
                for i, agent in enumerate(self.agents)
            },
            {agent: np.random.randint(0, 5) for agent in self.agents},
            {agent: 1 for agent in self.agents},
            {agent: np.random.randint(0, 2) for agent in self.agents},
            self.reset()[1],
        )


class MultiAgentCNNActor(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(
            in_channels=3, out_channels=16, kernel_size=(3, 3), stride=4
        )
        self.conv2 = nn.Conv2d(
            in_channels=16, out_channels=32, kernel_size=(3, 3), stride=2
        )
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.fc1 = nn.Linear(288, 256)
        self.fc2 = nn.Linear(256, 2)
        self.relu = nn.ReLU()
        self.output_activation = GumbelSoftmax()

    def forward(self, state_tensor):
        x = self.relu(self.conv1(state_tensor))
        x = self.relu(self.conv2(x))
        x = x.view(x.size(0), -1)
        x = self.relu(self.fc1(x))
        action_dist = self.output_activation(self.fc2(x))
        return action_dist


class MultiAgentCNNCritic(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(
            in_channels=3, out_channels=16, kernel_size=(3, 3), stride=4
        )
        self.conv2 = nn.Conv2d(
            in_channels=16, out_channels=32, kernel_size=(3, 3), stride=2
        )
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.fc1 = nn.Linear(288, 256)
        self.fc2 = nn.Linear(256, 1)
        self.relu = nn.ReLU()

    def forward(self, state_tensor):
        x = self.relu(self.conv1(state_tensor))
        x = self.relu(self.conv2(x))
        x = x.view(x.size(0), -1)
        x = self.relu(self.fc1(x))
        return self.fc2(x)


class DummyStochasticActor(StochasticActor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, *args, **kwargs):
        return super().forward(*args, **kwargs)

    def no_sync(self):
        class DummyNoSync:
            def __enter__(self):
                return self

            def __exit__(self, exc_type, exc_value, traceback):
                pass  # Add cleanup or handling if needed

        return DummyNoSync()


class DummyValueNetwork(ValueNetwork):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, *args, **kwargs):
        return super().forward(*args, **kwargs)

    def no_sync(self):
        class DummyNoSync:
            def __enter__(self):
                return self

            def __exit__(self, exc_type, exc_value, traceback):
                pass  # Add cleanup or handling if needed

        return DummyNoSync()


@pytest.fixture(scope="function")
def mlp_actor(observation_spaces, action_spaces, request):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    return nn.Sequential(
        nn.Linear(observation_spaces[0].shape[0], 64),
        nn.ReLU(),
        nn.Linear(64, action_spaces[0].n),
        nn.Softmax(dim=-1),
    )


@pytest.fixture(scope="function")
def mlp_critic(observation_spaces, request):
    observation_spaces = request.getfixturevalue(observation_spaces)
    return nn.Sequential(
        nn.Linear(observation_spaces[0].shape[0], 64),
        nn.ReLU(),
        nn.Linear(64, 1),
    )


@pytest.fixture(scope="function")
def cnn_actor():
    return MultiAgentCNNActor()


@pytest.fixture(scope="function")
def cnn_critic():
    return MultiAgentCNNCritic()


@pytest.fixture(scope="module")
def mocked_accelerator():
    MagicMock(spec=Accelerator)


@pytest.fixture(scope="function")
def accelerated_experiences(
    batch_size, observation_spaces, action_spaces, agent_ids, device, request
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    one_hot = all(isinstance(space, Discrete) for space in observation_spaces)
    discrete_actions = all(isinstance(space, Discrete) for space in action_spaces)
    state_size = (
        observation_spaces[0].shape if not one_hot else (observation_spaces[0].n,)
    )
    action_size = 1 if discrete_actions else action_spaces[0].shape[0]
    if one_hot:
        states = {
            agent: np.random.randint(0, state_size[0], (batch_size, 1))
            for agent in agent_ids
        }
    else:
        states = {
            agent: np.random.randn(batch_size, *state_size) for agent in agent_ids
        }

    if discrete_actions:
        actions = {
            agent: np.random.randint(0, action_size, (batch_size,))
            for agent in agent_ids
        }
    else:
        actions = {
            agent: np.random.randn(batch_size, action_size) for agent in agent_ids
        }
    log_probs = {agent: np.random.randn(batch_size, 1) for agent in agent_ids}
    rewards = {agent: np.random.randn(batch_size, 1) for agent in agent_ids}
    dones = {agent: np.random.randint(0, 2, (batch_size, 1)) for agent in agent_ids}
    values = {agent: np.random.randn(batch_size, 1) for agent in agent_ids}
    if one_hot:
        next_state = {
            agent: np.random.randint(0, state_size[0], (1,)) for agent in agent_ids
        }
    else:
        next_state = {agent: np.random.randn(*state_size) for agent in agent_ids}

    next_done = {agent: np.random.randint(0, 2, (1,)) for agent in agent_ids}

    return states, actions, log_probs, rewards, dones, values, next_state, next_done


@pytest.fixture(scope="function")
def experiences(
    batch_size, observation_spaces, action_spaces, agent_ids, device, request
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    one_hot = all(isinstance(space, Discrete) for space in observation_spaces)
    discrete_actions = all(isinstance(space, Discrete) for space in action_spaces)
    state_size = (
        observation_spaces[0].shape if not one_hot else (observation_spaces[0].n,)
    )
    action_size = 1 if discrete_actions else action_spaces[0].shape[0]
    if one_hot:
        states = {
            agent: np.random.randint(0, state_size[0], (batch_size, 1))
            for agent in agent_ids
        }
    else:
        states = {
            agent: np.random.randn(batch_size, *state_size) for agent in agent_ids
        }

    if discrete_actions:
        actions = {
            agent: np.random.randint(0, action_size, (batch_size, 1))
            for agent in agent_ids
        }
    else:
        actions = {
            agent: np.random.randn(batch_size, action_size) for agent in agent_ids
        }
    log_probs = {agent: np.random.randn(batch_size, 1) for agent in agent_ids}
    rewards = {agent: np.random.randn(batch_size) for agent in agent_ids}
    dones = {agent: np.random.randint(0, 2, (batch_size,)) for agent in agent_ids}
    values = {agent: np.random.randn(batch_size, 1) for agent in agent_ids}
    if one_hot:
        next_state = {
            agent: np.random.randint(0, state_size[0], (1,)) for agent in agent_ids
        }
    else:
        next_state = {agent: np.random.randn(*state_size) for agent in agent_ids}

    next_done = {agent: np.random.randint(0, 2, (1,)) for agent in agent_ids}

    return states, actions, log_probs, rewards, dones, values, next_state, next_done


@pytest.fixture(scope="function")
def vectorized_experiences(
    batch_size, vect_dim, observation_spaces, action_spaces, agent_ids, device, request
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    one_hot = all(isinstance(space, Discrete) for space in observation_spaces)
    discrete_actions = all(isinstance(space, Discrete) for space in action_spaces)
    state_size = (
        observation_spaces[0].shape if not one_hot else (observation_spaces[0].n,)
    )
    action_size = 1 if discrete_actions else action_spaces[0].shape[0]
    if one_hot:
        states = {
            agent: np.random.randint(0, state_size[0], (batch_size, vect_dim, 1))
            for agent in agent_ids
        }
    else:
        states = {
            agent: np.random.randn(batch_size, vect_dim, *state_size)
            for agent in agent_ids
        }

    if discrete_actions:
        actions = {
            agent: np.random.randint(0, action_size, (batch_size, vect_dim, 1))
            for agent in agent_ids
        }
    else:
        actions = {
            agent: np.random.randn(batch_size, vect_dim, action_size)
            for agent in agent_ids
        }
    log_probs = {agent: np.random.randn(batch_size, vect_dim, 1) for agent in agent_ids}
    rewards = {agent: np.random.randn(batch_size, vect_dim) for agent in agent_ids}
    dones = {
        agent: np.random.randint(0, 2, (batch_size, vect_dim)) for agent in agent_ids
    }
    values = {agent: np.random.randn(batch_size, vect_dim, 1) for agent in agent_ids}
    if one_hot:
        next_state = {
            agent: np.random.randint(
                0,
                state_size[0],
                (
                    vect_dim,
                    1,
                ),
            )
            for agent in agent_ids
        }
    else:
        next_state = {
            agent: np.random.randn(vect_dim, *state_size) for agent in agent_ids
        }

    next_done = {agent: np.random.randint(0, 2, (1, vect_dim)) for agent in agent_ids}

    return states, actions, log_probs, rewards, dones, values, next_state, next_done


@pytest.mark.parametrize("sum_score", [True, False])
@pytest.mark.parametrize("compile_mode", [None, "default"])
@pytest.mark.parametrize("observation_spaces", ["ma_vector_space", "ma_image_space"])
@pytest.mark.parametrize("vectorized", [False, True])
def test_loop(
    device,
    sum_score,
    compile_mode,
    observation_spaces,
    vectorized,
    ma_discrete_space,
    request,
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    if vectorized:
        env = make_multi_agent_vect_envs(
            DummyMultiEnv,
            2,
            **dict(
                observation_spaces=observation_spaces, action_spaces=ma_discrete_space
            ),
        )
    else:
        env = DummyMultiEnv(observation_spaces, ma_discrete_space)

    ippo = IPPO(
        observation_spaces,
        ma_discrete_space,
        agent_ids=["agent_0", "agent_1", "other_agent_0"],
        device=device,
        torch_compiler=compile_mode,
    )
    mean_score = ippo.test(env, max_steps=10, sum_scores=sum_score)
    if sum_score:
        assert isinstance(mean_score, float)
    else:
        assert isinstance(mean_score, np.ndarray)
        assert len(mean_score) == 2
    ippo.clean_up()


@pytest.mark.parametrize("observation_spaces", ["ma_vector_space"])
@pytest.mark.parametrize("compile_mode", [None, "default"])
@pytest.mark.parametrize("accelerator_flag", [False, True])
@pytest.mark.parametrize("wrap", [True, False])
def test_ippo_clone_returns_identical_agent(
    accelerator_flag, wrap, compile_mode, observation_spaces, ma_discrete_space, request
):
    # Clones the agent and returns an identical copy.
    observation_spaces = request.getfixturevalue(observation_spaces)
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    index = 0
    batch_size = 64
    lr = 1e-4
    learn_step = 2048
    gamma = 0.99
    gae_lambda = 0.95
    mut = None
    action_std_init = 0.0
    clip_coef = 0.2
    ent_coef = 0.01
    vf_coef = 0.5
    max_grad_norm = 0.5
    target_kl = None
    update_epochs = 4
    actor_networks = None
    critic_networks = None
    device = "cpu"
    accelerator = Accelerator(device_placement=False) if accelerator_flag else None

    ippo = IPPO(
        observation_spaces,
        ma_discrete_space,
        agent_ids,
        index=index,
        batch_size=batch_size,
        lr=lr,
        learn_step=learn_step,
        gamma=gamma,
        gae_lambda=gae_lambda,
        mut=mut,
        action_std_init=action_std_init,
        clip_coef=clip_coef,
        ent_coef=ent_coef,
        vf_coef=vf_coef,
        max_grad_norm=max_grad_norm,
        target_kl=target_kl,
        update_epochs=update_epochs,
        actor_networks=actor_networks,
        critic_networks=critic_networks,
        device=device,
        accelerator=accelerator,
        wrap=wrap,
        torch_compiler=compile_mode,
    )

    clone_agent = ippo.clone(wrap=wrap)

    assert isinstance(clone_agent, IPPO)
    assert clone_agent.observation_spaces == ippo.observation_spaces
    assert clone_agent.action_spaces == ippo.action_spaces
    assert clone_agent.n_agents == ippo.n_agents
    assert clone_agent.agent_ids == ippo.agent_ids
    assert clone_agent.shared_agent_ids == ippo.shared_agent_ids
    assert clone_agent.grouped_agents == ippo.grouped_agents
    assert clone_agent.unique_observation_spaces == ippo.unique_observation_spaces
    assert clone_agent.unique_action_spaces == ippo.unique_action_spaces
    assert clone_agent.grouped_spaces == ippo.grouped_spaces
    assert clone_agent.n_unique_agents == ippo.n_unique_agents
    assert clone_agent.index == ippo.index
    assert clone_agent.batch_size == ippo.batch_size
    assert clone_agent.lr == ippo.lr
    assert clone_agent.learn_step == ippo.learn_step
    assert clone_agent.gamma == ippo.gamma
    assert clone_agent.gae_lambda == ippo.gae_lambda
    assert clone_agent.action_std_init == ippo.action_std_init
    assert clone_agent.clip_coef == ippo.clip_coef
    assert clone_agent.ent_coef == ippo.ent_coef
    assert clone_agent.vf_coef == ippo.vf_coef
    assert clone_agent.max_grad_norm == ippo.max_grad_norm
    assert clone_agent.target_kl == ippo.target_kl
    assert clone_agent.update_epochs == ippo.update_epochs
    assert clone_agent.device == ippo.device
    assert clone_agent.accelerator == ippo.accelerator

    for shared_id in ippo.shared_agent_ids:
        actor = ippo.actors[shared_id]
        clone_actor = clone_agent.actors[shared_id]
        assert_state_dicts_equal(clone_actor.state_dict(), actor.state_dict())

        critic = ippo.critics[shared_id]
        clone_critic = clone_agent.critics[shared_id]
        assert_state_dicts_equal(clone_critic.state_dict(), critic.state_dict())


@pytest.mark.parametrize("compile_mode", [None])
def test_clone_new_index(compile_mode, ma_vector_space, ma_discrete_space):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]

    ippo = IPPO(
        ma_vector_space,
        ma_discrete_space,
        agent_ids,
        torch_compiler=compile_mode,
    )
    clone_agent = ippo.clone(index=100)

    assert clone_agent.index == 100
    ippo.clean_up()
    clone_agent.clean_up()


@pytest.mark.parametrize("compile_mode", [None])
def test_clone_after_learning(compile_mode, ma_vector_space, ma_discrete_space):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    batch_size = 8

    ippo = IPPO(
        ma_vector_space,
        ma_discrete_space,
        agent_ids,
        batch_size=batch_size,
        torch_compiler=compile_mode,
        target_kl=1e-10,
    )

    states = {
        agent_id: np.random.randn(batch_size, ma_vector_space[idx].shape[0])
        for idx, agent_id in enumerate(agent_ids)
    }
    actions = {
        agent_id: np.random.randint(0, ma_discrete_space[idx].n, (batch_size,))
        for idx, agent_id in enumerate(agent_ids)
    }
    log_probs = {agent_id: np.random.randn(batch_size, 1) for agent_id in agent_ids}
    rewards = {agent_id: np.random.randn(batch_size, 1) for agent_id in agent_ids}
    dones = {agent_id: torch.zeros(batch_size, 1) for agent_id in agent_ids}
    values = {agent_id: np.random.randn(batch_size, 1) for agent_id in agent_ids}
    next_state = {
        agent_id: np.random.randn(ma_vector_space[idx].shape[0])
        for idx, agent_id in enumerate(agent_ids)
    }
    next_done = {agent: np.random.randint(0, 2, (1,)) for agent in agent_ids}

    experiences = (
        states,
        actions,
        log_probs,
        rewards,
        dones,
        values,
        next_state,
        next_done,
    )
    ippo.learn(experiences)
    clone_agent = ippo.clone()

    assert isinstance(clone_agent, IPPO)
    assert clone_agent.observation_spaces == ippo.observation_spaces
    assert clone_agent.action_spaces == ippo.action_spaces
    assert clone_agent.n_agents == ippo.n_agents
    assert clone_agent.agent_ids == ippo.agent_ids
    assert clone_agent.index == ippo.index
    assert clone_agent.batch_size == ippo.batch_size
    assert clone_agent.lr == ippo.lr
    assert clone_agent.learn_step == ippo.learn_step
    assert clone_agent.gamma == ippo.gamma
    assert clone_agent.gae_lambda == ippo.gae_lambda
    assert clone_agent.device == ippo.device
    assert clone_agent.accelerator == ippo.accelerator

    for shared_id in ippo.shared_agent_ids:
        actor = ippo.actors[shared_id]
        clone_actor = clone_agent.actors[shared_id]
        assert_state_dicts_equal(clone_actor.state_dict(), actor.state_dict())

        critic = ippo.critics[shared_id]
        clone_critic = clone_agent.critics[shared_id]
        assert_state_dicts_equal(clone_critic.state_dict(), critic.state_dict())

        actor_opt = ippo.actor_optimizers[shared_id]
        clone_actor_opt = clone_agent.actor_optimizers[shared_id]
        assert str(clone_actor_opt) == str(actor_opt)

        critic_opt = ippo.critic_optimizers[shared_id]
        clone_critic_opt = clone_agent.critic_optimizers[shared_id]
        assert str(clone_critic_opt) == str(critic_opt)
    ippo.clean_up()
    clone_agent.clean_up()


@pytest.mark.parametrize("compile_mode", [None])
@pytest.mark.parametrize("batch_size", [16])
@pytest.mark.parametrize("agent_ids", [["agent_0", "agent_1", "other_agent_0"]])
@pytest.mark.parametrize("observation_spaces", ["ma_vector_space"])
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space"])
def test_ippo_learns_from_experiences_distributed(
    observation_spaces,
    action_spaces,
    agent_ids,
    accelerated_experiences,
    compile_mode,
    batch_size,
    request,
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    accelerator = Accelerator(device_placement=False)
    ippo = IPPO(
        observation_spaces=observation_spaces,
        action_spaces=action_spaces,
        agent_ids=agent_ids,
        accelerator=accelerator,
        torch_compiler=compile_mode,
    )

    for shared_id in ippo.shared_agent_ids:
        actor = ippo.actors[shared_id]
        critic = ippo.critics[shared_id]
        actor.no_sync = no_sync.__get__(actor)
        critic.no_sync = no_sync.__get__(critic)

    actors = ippo.actors
    actors_pre_learn_sd = {
        shared_id: copy.deepcopy(actor.state_dict())
        for shared_id, actor in ippo.actors.items()
    }
    critics = ippo.critics
    critics_pre_learn_sd = {
        shared_id: copy.deepcopy(critic.state_dict())
        for shared_id, critic in ippo.critics.items()
    }

    for _ in range(2):
        ippo.scores.append(0)
        loss = ippo.learn(accelerated_experiences)

    assert isinstance(loss, dict)
    for agent_id in ippo.shared_agent_ids:
        assert agent_id in loss

    for shared_id in ippo.shared_agent_ids:
        old_actor = actors[shared_id]
        updated_actor = ippo.actors[shared_id]
        assert old_actor == updated_actor

        old_critic = critics[shared_id]
        updated_critic = ippo.critics[shared_id]
        assert old_critic == updated_critic

        old_critic_state_dict = critics_pre_learn_sd[shared_id]
        assert_not_equal_state_dict(old_critic_state_dict, updated_critic.state_dict())

        old_actor_state_dict = actors_pre_learn_sd[shared_id]
        assert_not_equal_state_dict(old_actor_state_dict, updated_actor.state_dict())


@pytest.mark.parametrize("compile_mode", [None, "default"])
@pytest.mark.parametrize("observation_spaces", ["ma_image_space", "ma_vector_space"])
@pytest.mark.parametrize("agent_ids", [["agent_0", "agent_1", "other_agent_0"]])
@pytest.mark.parametrize("batch_size", [64])
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space"])
def test_ippo_learns_from_experiences(
    observation_spaces,
    agent_ids,
    action_spaces,
    experiences,
    batch_size,
    device,
    compile_mode,
    request,
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    ippo = IPPO(
        observation_spaces=observation_spaces,
        action_spaces=action_spaces,
        agent_ids=agent_ids,
        device=device,
        torch_compiler=compile_mode,
    )

    actors = ippo.actors
    actors_pre_learn_sd = {
        shared_id: copy.deepcopy(actor.state_dict())
        for shared_id, actor in ippo.actors.items()
    }
    critics = ippo.critics
    critics_pre_learn_sd = {
        shared_id: copy.deepcopy(critic.state_dict())
        for shared_id, critic in ippo.critics.items()
    }

    for _ in range(2):
        ippo.scores.append(0)
        loss = ippo.learn(experiences)

    assert isinstance(loss, dict)
    for shared_id in ippo.shared_agent_ids:
        assert shared_id in loss

        old_actor = actors[shared_id]
        updated_actor = ippo.actors[shared_id]
        assert old_actor == updated_actor

        old_actor_state_dict = actors_pre_learn_sd[shared_id]
        assert_not_equal_state_dict(old_actor_state_dict, updated_actor.state_dict())

        old_critic = critics[shared_id]
        updated_critic = ippo.critics[shared_id]
        assert old_critic == updated_critic

        old_critic_state_dict = critics_pre_learn_sd[shared_id]
        assert_not_equal_state_dict(old_critic_state_dict, updated_critic.state_dict())


@pytest.mark.parametrize("compile_mode", [None])
@pytest.mark.parametrize("vect_dim", [1, 8])
@pytest.mark.parametrize("batch_size", [16])
@pytest.mark.parametrize("agent_ids", [["agent_0", "agent_1", "other_agent_0"]])
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space", "ma_vector_space"])
@pytest.mark.parametrize("observation_spaces", ["ma_image_space", "ma_vector_space"])
def test_ippo_learns_from_vectorized_experiences(
    agent_ids,
    observation_spaces,
    action_spaces,
    vectorized_experiences,
    batch_size,
    device,
    compile_mode,
    request,
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    ippo = IPPO(
        observation_spaces=observation_spaces,
        action_spaces=action_spaces,
        agent_ids=agent_ids,
        device=device,
        torch_compiler=compile_mode,
        batch_size=batch_size,
        lr=0.1,
    )

    actors = ippo.actors
    actors_pre_learn_sd = {
        shared_id: copy.deepcopy(actor.state_dict())
        for shared_id, actor in ippo.actors.items()
    }
    critics = ippo.critics
    critics_pre_learn_sd = {
        shared_id: copy.deepcopy(critic.state_dict())
        for shared_id, critic in ippo.critics.items()
    }

    for _ in range(2):
        ippo.scores.append(0)
        loss = ippo.learn(vectorized_experiences)

    assert isinstance(loss, dict)
    for agent_id in ippo.shared_agent_ids:
        assert agent_id in loss

    for shared_id in ippo.shared_agent_ids:
        old_actor = actors[shared_id]
        updated_actor = ippo.actors[shared_id]
        assert old_actor == updated_actor

        old_actor_state_dict = actors_pre_learn_sd[shared_id]
        assert_not_equal_state_dict(old_actor_state_dict, updated_actor.state_dict())

        old_critic = critics[shared_id]
        updated_critic = ippo.critics[shared_id]
        assert old_critic == updated_critic

        old_critic_state_dict = critics_pre_learn_sd[shared_id]
        assert_not_equal_state_dict(old_critic_state_dict, updated_critic.state_dict())


def test_ippo_learns_from_hardcoded_vectorized_experiences_mlp(
    ma_vector_space,
    device,
):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    states = {
        agent: np.array(
            [
                [[1, 1, 1, 1, 1, 1], [2, 2, 2, 2, 2, 2], [3, 3, 3, 3, 3, 3]],
                [[4, 4, 4, 4, 4, 4], [5, 5, 5, 5, 5, 5], [6, 6, 6, 6, 6, 6]],
                [[7, 7, 7, 7, 7, 7], [8, 8, 8, 8, 8, 8], [9, 9, 9, 9, 9, 9]],
            ]
        )
        * i
        for i, agent in enumerate(agent_ids)
    }
    actions = {
        agent: np.array(
            [
                [[1, 1, 1, 1, 1, 1], [2, 2, 2, 2, 2, 2], [3, 3, 3, 3, 3, 3]],
                [[4, 4, 4, 4, 4, 4], [5, 5, 5, 5, 5, 5], [6, 6, 6, 6, 6, 6]],
                [[7, 7, 7, 7, 7, 7], [8, 8, 8, 8, 8, 8], [9, 9, 9, 9, 9, 9]],
            ]
        )
        * i
        for i, agent in enumerate(agent_ids)
    }
    log_probs = {
        agent: np.array([[[1], [2], [3]], [[4], [5], [6]], [[7], [8], [9]]]) * i
        for i, agent in enumerate(agent_ids)
    }
    rewards = {
        agent: np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]) * i
        for i, agent in enumerate(agent_ids)
    }
    dones = {
        agent: np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        for i, agent in enumerate(agent_ids)
    }
    values = {
        agent: np.array([[[1], [2], [3]], [[4], [5], [6]], [[7], [8], [9]]]) * i
        for i, agent in enumerate(agent_ids)
    }
    next_state = {
        agent: np.array(
            [[4, 4, 4, 4, 4, 4], [7, 7, 7, 7, 7, 7], [10, 10, 10, 10, 10, 10]]
        )
        * i
        for i, agent in enumerate(agent_ids)
    }
    next_done = {agent: np.array([[0, 1, 0]]) for i, agent in enumerate(agent_ids)}

    experiences = (
        states,
        actions,
        log_probs,
        rewards,
        dones,
        values,
        next_state,
        next_done,
    )
    ippo = IPPO(
        observation_spaces=ma_vector_space,
        action_spaces=ma_vector_space,
        agent_ids=agent_ids,
        device=device,
        torch_compiler=None,
        batch_size=32,
    )

    actors = ippo.actors
    actors_pre_learn_sd = {
        shared_id: copy.deepcopy(actor.state_dict())
        for shared_id, actor in ippo.actors.items()
    }
    critics = ippo.critics
    critics_pre_learn_sd = {
        shared_id: copy.deepcopy(critic.state_dict())
        for shared_id, critic in ippo.critics.items()
    }

    for _ in range(4):
        ippo.scores.append(0)
        loss = ippo.learn(experiences)

    assert isinstance(loss, dict)
    for agent_id in ippo.shared_agent_ids:
        assert agent_id in loss

    for shared_id in ippo.shared_agent_ids:
        old_actor = actors[shared_id]
        updated_actor = ippo.actors[shared_id]
        assert old_actor == updated_actor

        old_actor_state_dict = actors_pre_learn_sd[shared_id]
        assert_not_equal_state_dict(old_actor_state_dict, updated_actor.state_dict())

        old_critic = critics[shared_id]
        updated_critic = ippo.critics[shared_id]
        assert old_critic == updated_critic

        old_critic_state_dict = critics_pre_learn_sd[shared_id]
        assert_not_equal_state_dict(old_critic_state_dict, updated_critic.state_dict())
    ippo.clean_up()


def no_sync(self):
    class DummyNoSync:
        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_value, traceback):
            pass  # Add cleanup or handling if needed

    return DummyNoSync()


@pytest.mark.parametrize(
    "action_spaces",
    [
        "ma_discrete_space",
        "ma_vector_space",
    ],
)
@pytest.mark.parametrize("compile_mode", [None])
def test_ippo_get_action_agent_masking(
    ma_vector_space, action_spaces, device, compile_mode, request
):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    state = {agent: np.random.randn(*ma_vector_space[0].shape) for agent in agent_ids}
    action_spaces = request.getfixturevalue(action_spaces)
    discrete_actions = all(
        isinstance(space, spaces.Discrete) for space in action_spaces
    )
    if discrete_actions:
        info = {
            "agent_0": {"env_defined_actions": 1},
            "agent_1": {"env_defined_actions": None},
            "other_agent_0": {"env_defined_actions": None},
        }
    else:
        info = {
            "agent_0": {"env_defined_actions": np.array([0, 1, 0, 1, 0, 1])},
            "agent_1": {"env_defined_actions": None},
            "other_agent_0": {"env_defined_actions": None},
        }

    # Define the ippo agent
    ippo = IPPO(
        observation_spaces=ma_vector_space,
        action_spaces=action_spaces,
        agent_ids=agent_ids,
        device=device,
        torch_compiler=compile_mode,
    )

    # Get the action
    actions, _, _, _ = ippo.get_action(obs=state, infos=info)

    if discrete_actions:
        assert np.array_equal(actions["agent_0"], np.array([[1]])), actions["agent_0"]
    else:
        assert np.array_equal(
            actions["agent_0"], np.array([[0, 1, 0, 1, 0, 1]])
        ), actions["agent_0"]
    ippo.clean_up()


@pytest.mark.parametrize(
    "observation_spaces",
    [
        "ma_discrete_space",
        "ma_vector_space",
        "ma_image_space",
        "ma_dict_space",
    ],
)
@pytest.mark.parametrize(
    "action_spaces",
    [
        "ma_discrete_space",
        "ma_vector_space",
        "ma_multidiscrete_space",
        "ma_multibinary_space",
    ],
)
@pytest.mark.parametrize("action_batch_size", [None, 16])
@pytest.mark.parametrize("compile_mode", [None, "default"])
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
def test_ippo_get_action(
    observation_spaces,
    action_spaces,
    device,
    compile_mode,
    accelerator,
    action_batch_size,
    request,
):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)

    state = {}
    for agent_id, obs_space in zip(agent_ids, observation_spaces):
        sample_state = get_sample_from_space(obs_space)

        if (
            not isinstance(obs_space, (spaces.Dict, spaces.Tuple))
            and action_batch_size is not None
        ):
            sample_state = np.stack([sample_state] * action_batch_size)
        else:
            action_batch_size = None

        state[agent_id] = sample_state

    info = None
    ippo = IPPO(
        observation_spaces=observation_spaces,
        action_spaces=action_spaces,
        agent_ids=agent_ids,
        device=device,
        torch_compiler=compile_mode,
        accelerator=accelerator,
        action_batch_size=action_batch_size,
    )
    actions, log_probs, dist_entropy, state_values = ippo.get_action(
        obs=state, infos=info
    )

    # Check action shapes
    assert isinstance(actions, dict)
    assert isinstance(log_probs, dict)

    for agent_id in agent_ids:
        assert agent_id in actions
        assert agent_id in log_probs
        assert agent_id in dist_entropy
        assert agent_id in state_values
    ippo.clean_up()


@pytest.mark.parametrize(
    "observation_spaces",
    [
        "ma_vector_space",
        "ma_image_space",
    ],
)
@pytest.mark.parametrize(
    "action_spaces",
    [
        "ma_vector_space",
        "ma_discrete_space",
    ],
)
@pytest.mark.parametrize("compile_mode", [None])
def test_ippo_get_action_vectorized(
    observation_spaces, action_spaces, device, compile_mode, request
):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    vect_dim = 2
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    if all(isinstance(space, spaces.Discrete) for space in observation_spaces):
        state = {
            agent: np.random.randint(0, observation_spaces[idx].n, (vect_dim, 1))
            for idx, agent in enumerate(agent_ids)
        }
        info = None
    else:
        state = {
            agent: np.random.randn(vect_dim, *observation_spaces[idx].shape).astype(
                np.float32
            )
            for idx, agent in enumerate(agent_ids)
        }
        info = {agent: {} for agent in agent_ids}

    ippo = IPPO(
        observation_spaces=observation_spaces,
        action_spaces=action_spaces,
        agent_ids=agent_ids,
        device=device,
        torch_compiler=compile_mode,
    )
    actions, log_probs, dist_entropy, state_values = ippo.get_action(
        obs=state, infos=info
    )

    # Check action shapes
    assert isinstance(actions, dict)
    assert isinstance(log_probs, dict)

    for agent_id in agent_ids:
        assert agent_id in actions
        assert agent_id in log_probs
        assert agent_id in dist_entropy
        assert agent_id in state_values


def test_ippo_get_action_action_masking_exception(
    ma_vector_space, ma_discrete_space, device
):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    state = {
        agent: {
            "observation": np.random.randn(*ma_vector_space[idx].shape),
            "action_mask": [0, 1, 0, 1],
        }
        for idx, agent in enumerate(agent_ids)
    }
    ippo = IPPO(
        observation_spaces=ma_vector_space,
        action_spaces=ma_discrete_space,
        agent_ids=agent_ids,
        device=device,
    )
    with pytest.raises(AssertionError):
        actions, log_probs, dist_entropy, state_values = ippo.get_action(obs=state)
    ippo.clean_up()


def test_ippo_get_action_action_masking(ma_vector_space, ma_discrete_space, device):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    state = {
        agent: np.random.randn(*ma_vector_space[idx].shape).astype(np.float32)
        for idx, agent in enumerate(agent_ids)
    }
    info = {
        agent: {
            "action_mask": [0, 1],
        }
        for agent in agent_ids
    }
    ippo = IPPO(
        observation_spaces=ma_vector_space,
        action_spaces=ma_discrete_space,
        agent_ids=agent_ids,
        device=device,
    )
    actions, _, _, _ = ippo.get_action(obs=state, infos=info)

    # With action mask [0, 1, 0, 1], actions should be index 1 or 3
    for agent_id, action_array in actions.items():
        for action in action_array:
            assert action in [1, 3]
    ippo.clean_up()


@pytest.mark.parametrize(
    "mode", (None, 0, False, "default", "reduce-overhead", "max-autotune")
)
def test_ippo_init_torch_compiler_no_error(ma_vector_space, ma_discrete_space, mode):
    ippo = IPPO(
        observation_spaces=ma_vector_space,
        action_spaces=ma_discrete_space,
        agent_ids=["agent_0", "agent_1", "other_agent_0"],
        device="cuda" if torch.cuda.is_available() else "cpu",
        torch_compiler=mode,
    )
    if isinstance(mode, str):
        assert all(
            isinstance(a, torch._dynamo.eval_frame.OptimizedModule)
            for a in ippo.actors.values()
        )
        assert all(
            isinstance(a, torch._dynamo.eval_frame.OptimizedModule)
            for a in ippo.critics.values()
        )
        assert ippo.torch_compiler == mode
    else:
        assert isinstance(ippo, IPPO)
    ippo.clean_up()


@pytest.mark.parametrize("mode", (1, True, "max-autotune-no-cudagraphs"))
def test_ippo_init_torch_compiler_error(
    mode, ma_vector_space, ma_discrete_space, device
):
    err_string = (
        "Choose between torch compiler modes: "
        "default, reduce-overhead, max-autotune or None"
    )
    with pytest.raises(AssertionError, match=err_string):
        IPPO(
            observation_spaces=ma_vector_space,
            action_spaces=ma_discrete_space,
            agent_ids=["agent_0", "agent_1", "other_agent_0"],
            device="cuda" if torch.cuda.is_available() else "cpu",
            torch_compiler=mode,
        )


@pytest.mark.parametrize(
    "observation_spaces",
    [
        "ma_vector_space",
        "ma_image_space",
    ],
)
@pytest.mark.parametrize("accelerator_flag", [False, True])
@pytest.mark.parametrize("compile_mode", [None])
def test_initialize_ippo_with_net_config(
    accelerator_flag,
    observation_spaces,
    ma_discrete_space,
    device,
    compile_mode,
    request,
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    batch_size = 64
    accelerator = Accelerator() if accelerator_flag else None

    net_config = {
        "encoder_config": get_default_encoder_config(observation_spaces[0]),
        "head_config": {"hidden_size": [32]},
    }
    ippo = IPPO(
        observation_spaces=observation_spaces,
        action_spaces=ma_discrete_space,
        agent_ids=agent_ids,
        net_config=net_config,
        accelerator=accelerator,
        device=device,
        torch_compiler=compile_mode,
        target_kl=0.5,
    )

    assert ippo.observation_spaces == observation_spaces
    assert ippo.action_spaces == ma_discrete_space
    assert ippo.n_agents == len(agent_ids)
    assert ippo.agent_ids == agent_ids
    assert ippo.batch_size == batch_size
    assert ippo.scores == []
    assert ippo.fitness == []
    assert ippo.steps == [0]
    assert ippo.target_kl == 0.5

    expected_actor_cls = (
        OptimizedModule
        if compile_mode is not None and accelerator is None
        else StochasticActor
    )
    expected_critic_cls = (
        OptimizedModule
        if compile_mode is not None and accelerator is None
        else ValueNetwork
    )
    assert all(isinstance(actor, expected_actor_cls) for actor in ippo.actors.values())
    assert all(
        isinstance(critic, expected_critic_cls) for critic in ippo.critics.values()
    )

    expected_opt_cls = optim.Adam if accelerator is None else AcceleratedOptimizer
    assert all(
        isinstance(actor_optimizer, expected_opt_cls)
        for actor_optimizer in ippo.actor_optimizers.values()
    )
    assert all(
        isinstance(critic_optimizer, expected_opt_cls)
        for critic_optimizer in ippo.critic_optimizers.values()
    )
    assert isinstance(ippo.criterion, nn.MSELoss)
    ippo.clean_up()


# TODO: This will be deprecated in the future
@pytest.mark.parametrize("accelerator_flag", [False, True])
@pytest.mark.parametrize("compile_mode", [None])
@pytest.mark.parametrize("observation_spaces", ["ma_vector_space"])
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space"])
def test_initialize_ippo_with_mlp_networks(
    mlp_actor,
    mlp_critic,
    observation_spaces,
    action_spaces,
    accelerator_flag,
    device,
    compile_mode,
    request,
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    accelerator = Accelerator() if accelerator_flag else None
    evo_actors = ModuleDict(
        {
            "agent": MakeEvolvable(
                network=mlp_actor, input_tensor=torch.randn(1, 6), device=device
            ),
            "other_agent": MakeEvolvable(
                network=mlp_actor, input_tensor=torch.randn(1, 6), device=device
            ),
        }
    )
    evo_critics = ModuleDict(
        {
            "agent": MakeEvolvable(
                network=mlp_critic, input_tensor=torch.randn(1, 6), device=device
            ),
            "other_agent": MakeEvolvable(
                network=mlp_critic, input_tensor=torch.randn(1, 6), device=device
            ),
        }
    )
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    ippo = IPPO(
        observation_spaces=observation_spaces,
        action_spaces=action_spaces,
        agent_ids=agent_ids,
        actor_networks=evo_actors,
        critic_networks=evo_critics,
        device=device,
        accelerator=accelerator,
        torch_compiler=compile_mode,
    )
    if compile_mode is not None and accelerator is None:
        assert all(isinstance(actor, OptimizedModule) for actor in ippo.actors.values())
        assert all(
            isinstance(critic, OptimizedModule) for critic in ippo.critics.values()
        )
    else:
        assert all(isinstance(actor, MakeEvolvable) for actor in ippo.actors.values())
        assert all(
            isinstance(critic, MakeEvolvable) for critic in ippo.critics.values()
        )

    assert ippo.observation_spaces == observation_spaces
    assert ippo.action_spaces == action_spaces
    assert ippo.n_agents == 3
    assert ippo.agent_ids == agent_ids
    assert ippo.scores == []
    assert ippo.fitness == []
    assert ippo.steps == [0]
    if accelerator is None:
        assert all(
            isinstance(actor_optimizer, optim.Adam)
            for actor_optimizer in ippo.actor_optimizers.values()
        )
        assert all(
            isinstance(critic_optimizer, optim.Adam)
            for critic_optimizer in ippo.critic_optimizers.values()
        )
    else:
        assert all(
            isinstance(actor_optimizer, AcceleratedOptimizer)
            for actor_optimizer in ippo.actor_optimizers.values()
        )
        assert all(
            isinstance(critic_optimizer, AcceleratedOptimizer)
            for critic_optimizer in ippo.critic_optimizers.values()
        )
    assert isinstance(ippo.criterion, nn.MSELoss)


@pytest.mark.parametrize("observation_spaces", ["ma_vector_space"])
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space"])
def test_initialize_ippo_with_mlp_networks_gumbel_softmax(
    mlp_actor,
    mlp_critic,
    observation_spaces,
    action_spaces,
    device,
    request,
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    net_config = {
        "encoder_config": {
            "hidden_size": [64, 64],
            "min_hidden_layers": 1,
            "max_hidden_layers": 3,
            "min_mlp_nodes": 64,
            "max_mlp_nodes": 500,
            "activation": "ReLU",
            "init_layers": False,
        },
        "head_config": {
            "output_activation": "GumbelSoftmax",
            "activation": "ReLU",
            "hidden_size": [64, 64],
            "init_layers": False,
        },
    }
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    ippo = IPPO(
        observation_spaces=observation_spaces,
        action_spaces=action_spaces,
        agent_ids=agent_ids,
        net_config=net_config,
        device=device,
        torch_compiler="reduce-overhead",
    )
    assert ippo.torch_compiler == "reduce-overhead"
    ippo.clean_up()


# TODO: This will be deprecated in the future
@pytest.mark.parametrize("accelerator_flag", [False, True])
@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_initialize_ippo_with_cnn_networks(
    cnn_actor,
    cnn_critic,
    ma_image_space,
    ma_discrete_space,
    accelerator_flag,
    device,
    compile_mode,
):
    if accelerator_flag:
        accelerator = Accelerator()
    else:
        accelerator = None
    evo_actors = ModuleDict(
        {
            "agent": MakeEvolvable(
                network=cnn_actor,
                input_tensor=torch.randn(1, *ma_image_space[0].shape),
                device=device,
            ),
            "other_agent": MakeEvolvable(
                network=cnn_actor,
                input_tensor=torch.randn(1, *ma_image_space[0].shape),
                device=device,
            ),
        }
    )
    evo_critics = ModuleDict(
        {
            "agent": MakeEvolvable(
                network=cnn_critic,
                input_tensor=torch.randn(1, *ma_image_space[0].shape),
                device=device,
            ),
            "other_agent": MakeEvolvable(
                network=cnn_critic,
                input_tensor=torch.randn(1, *ma_image_space[0].shape),
                device=device,
            ),
        }
    )
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    ippo = IPPO(
        observation_spaces=ma_image_space,
        action_spaces=ma_discrete_space,
        agent_ids=agent_ids,
        actor_networks=evo_actors,
        critic_networks=evo_critics,
        device=device,
        accelerator=accelerator,
        torch_compiler=compile_mode,
    )
    if compile_mode is not None and accelerator is None:
        assert all(isinstance(actor, OptimizedModule) for actor in ippo.actors.values())
        assert all(
            isinstance(critic, OptimizedModule) for critic in ippo.critics.values()
        )
    else:
        assert all(isinstance(actor, MakeEvolvable) for actor in ippo.actors.values())
        assert all(
            isinstance(critic, MakeEvolvable) for critic in ippo.critics.values()
        )

    assert ippo.observation_spaces == ma_image_space
    assert ippo.action_spaces == ma_discrete_space
    assert ippo.n_agents == 3
    assert ippo.agent_ids == agent_ids
    assert ippo.scores == []
    assert ippo.fitness == []
    assert ippo.steps == [0]
    if accelerator is None:
        assert all(
            isinstance(actor_optimizer, optim.Adam)
            for actor_optimizer in ippo.actor_optimizers.values()
        )
        assert all(
            isinstance(critic_optimizer, optim.Adam)
            for critic_optimizer in ippo.critic_optimizers.values()
        )
    else:
        assert all(
            isinstance(actor_optimizer, AcceleratedOptimizer)
            for actor_optimizer in ippo.actor_optimizers.values()
        )
        assert all(
            isinstance(critic_optimizer, AcceleratedOptimizer)
            for critic_optimizer in ippo.critic_optimizers.values()
        )
    assert isinstance(ippo.criterion, nn.MSELoss)


@pytest.mark.parametrize(
    "observation_spaces, net",
    [
        ("ma_image_space", "cnn"),
        ("ma_vector_space", "mlp"),
    ],
)
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_initialize_ippo_with_evo_networks(
    observation_spaces,
    ma_discrete_space,
    net,
    device,
    compile_mode,
    accelerator,
    request,
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    net_config = get_default_encoder_config(observation_spaces[0])

    head_config = {
        "output_activation": "Softmax" if net == "mlp" else "GumbelSoftmax",
        "activation": "ReLU",
        "hidden_size": [64, 64],
        "init_layers": False,
    }

    critic_head_config = copy.deepcopy(head_config)
    critic_head_config.update({"output_activation": None})

    net_config = {"encoder_config": net_config, "head_config": head_config}
    critic_net_config = {
        "encoder_config": copy.deepcopy(net_config["encoder_config"]),
        "head_config": critic_head_config,
    }

    evo_actors = ModuleDict(
        {
            "agent": StochasticActor(
                observation_spaces[0], ma_discrete_space[0], device=device, **net_config
            ),
            "other_agent": StochasticActor(
                observation_spaces[2], ma_discrete_space[2], device=device, **net_config
            ),
        }
    )
    evo_critics = ModuleDict(
        {
            "agent": ValueNetwork(
                observation_space=observation_spaces[0],
                device=device,
                **critic_net_config,
            ),
            "other_agent": ValueNetwork(
                observation_space=observation_spaces[1],
                device=device,
                **critic_net_config,
            ),
        }
    )
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    ippo = IPPO(
        observation_spaces=observation_spaces,
        action_spaces=ma_discrete_space,
        agent_ids=agent_ids,
        actor_networks=evo_actors,
        critic_networks=evo_critics,
        device=device,
        torch_compiler=compile_mode,
        accelerator=accelerator,
    )

    expected_actor_cls = (
        OptimizedModule
        if compile_mode is not None and accelerator is None
        else StochasticActor
    )
    expected_critic_cls = (
        OptimizedModule
        if compile_mode is not None and accelerator is None
        else ValueNetwork
    )
    assert all(isinstance(actor, expected_actor_cls) for actor in ippo.actors.values())
    assert all(
        isinstance(critic, expected_critic_cls) for critic in ippo.critics.values()
    )

    expected_opt_cls = optim.Adam if accelerator is None else AcceleratedOptimizer
    assert all(
        isinstance(actor_optimizer, expected_opt_cls)
        for actor_optimizer in ippo.actor_optimizers.values()
    )
    assert all(
        isinstance(critic_optimizer, expected_opt_cls)
        for critic_optimizer in ippo.critic_optimizers.values()
    )

    assert ippo.observation_spaces == observation_spaces
    assert ippo.action_spaces == ma_discrete_space
    assert ippo.n_agents == 3
    assert ippo.agent_ids == agent_ids
    assert ippo.scores == []
    assert ippo.fitness == []
    assert ippo.steps == [0]
    assert isinstance(ippo.criterion, nn.MSELoss)
    ippo.clean_up()


@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_initialize_ippo_with_incorrect_evo_networks(
    compile_mode, ma_vector_space, ma_discrete_space
):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    evo_actors = []
    evo_critics = []

    with pytest.raises(AssertionError):
        ippo = IPPO(
            observation_spaces=ma_vector_space,
            action_spaces=ma_discrete_space,
            agent_ids=agent_ids,
            actor_networks=evo_actors,
            critic_networks=evo_critics,
            torch_compiler=compile_mode,
        )
        assert ippo


@pytest.mark.parametrize(
    "observation_spaces, action_spaces, actors, critics",
    [
        (
            "ma_vector_space",
            "ma_discrete_space",
            [EvolvableMLP(4, 2, [32]) for _ in range(2)],
            [1 for _ in range(2)],
        ),
        (
            "ma_vector_space",
            "ma_discrete_space",
            [1 for _ in range(2)],
            [EvolvableMLP(4, 2, [32]) for _ in range(2)],
        ),
    ],
)
def test_initialize_ippo_with_incorrect_networks(
    observation_spaces, action_spaces, actors, critics, request
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    with pytest.raises(TypeError):
        ippo = IPPO(
            observation_spaces=observation_spaces,
            action_spaces=action_spaces,
            agent_ids=["agent_0", "agent_1", "other_agent_0"],
            actor_networks=actors,
            critic_networks=critics,
        )
        assert ippo


@pytest.mark.parametrize("observation_spaces", ["ma_vector_space"])
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space"])
def test_ippo_init_warning(
    mlp_actor, observation_spaces, action_spaces, device, request
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    warning_string = "Actor and critic network lists must both be supplied to use custom networks. Defaulting to net config."
    evo_actors = [
        MakeEvolvable(network=mlp_actor, input_tensor=torch.randn(1, 6), device=device)
        for _ in range(2)
    ]
    with pytest.warns(UserWarning, match=warning_string):
        IPPO(
            observation_spaces=observation_spaces,
            action_spaces=action_spaces,
            agent_ids=["agent_0", "agent_1", "other_agent_0"],
            actor_networks=evo_actors,
            device=device,
        )


def test_grouped_outputs_functions(ma_vector_space, ma_discrete_space):
    """Test that the assemble_grouped_outputs and disassemble_grouped_outputs
    functions work as expected and are inverses of each other."""

    # Initialize agent with grouped agents
    compile_mode = None
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    agent = IPPO(
        observation_spaces=ma_vector_space,
        action_spaces=ma_discrete_space,
        agent_ids=agent_ids,
        device="cpu",
        torch_compiler=compile_mode,
    )

    # Setting up grouped agent configuration
    agent.shared_agent_ids = ["agent", "other_agent"]
    agent.grouped_agents = {
        "agent": ["agent_0", "agent_1"],
        "other_agent": ["other_agent_0"],
    }

    # Create sample individual agent outputs
    vect_dim = 4
    output_dim = 5
    agent_outputs = {
        "agent_0": np.random.rand(vect_dim, output_dim),
        "agent_1": np.random.rand(vect_dim, output_dim),
        "other_agent_0": np.random.rand(vect_dim, output_dim),
    }

    # Test assemble_grouped_outputs
    grouped_outputs = agent.assemble_grouped_outputs(agent_outputs, vect_dim)

    # Check that the grouped outputs have the correct keys
    assert set(grouped_outputs.keys()) == {"agent", "other_agent"}

    # Check that the agent outputs are assembled correctly
    assert grouped_outputs["agent"].shape == (2 * vect_dim, output_dim)
    assert grouped_outputs["other_agent"].shape == (1 * vect_dim, output_dim)

    # Test disassemble_grouped_outputs
    disassembled_outputs = agent.disassemble_grouped_outputs(
        grouped_outputs, vect_dim, agent.grouped_agents
    )

    # Check that the disassembled outputs have the correct keys
    assert set(disassembled_outputs.keys()) == {"agent_0", "agent_1", "other_agent_0"}

    # Check that the outputs are correctly disassembled
    for agent_id in agent_ids:
        assert disassembled_outputs[agent_id].shape == (vect_dim, output_dim)

    # Check the round trip consistency by comparing the original and disassembled outputs
    for agent_id in agent_ids:
        np.testing.assert_allclose(
            agent_outputs[agent_id],
            disassembled_outputs[agent_id],
            rtol=1e-5,
            atol=1e-5,
        )
    agent.clean_up()


@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_get_action_distributed(compile_mode, ma_vector_space, ma_discrete_space):
    accelerator = Accelerator()
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    state = {
        agent: np.random.randn(*ma_vector_space[idx].shape).astype(np.float32)
        for idx, agent in enumerate(agent_ids)
    }
    ippo = IPPO(
        observation_spaces=ma_vector_space,
        action_spaces=ma_discrete_space,
        agent_ids=agent_ids,
        accelerator=accelerator,
        torch_compiler=compile_mode,
        net_config={
            "encoder_config": {"hidden_size": [16, 16], "init_layers": False},
            "head_config": {"hidden_size": [16], "init_layers": False},
        },
    )
    new_actors = ModuleDict(
        {
            shared_id: DummyStochasticActor(
                observation_space=actor.observation_space,
                action_space=actor.action_space,
                device=actor.device,
                action_std_init=ippo.action_std_init,
                encoder_config={"hidden_size": [16, 16], "init_layers": False},
                head_config={"hidden_size": [16], "init_layers": False},
            )
            for shared_id, actor in ippo.actors.items()
        }
    )
    ippo.actors = new_actors
    actions, log_probs, dist_entropy, state_values = ippo.get_action(obs=state)

    # Check returns are the proper format
    assert isinstance(actions, dict)
    assert isinstance(log_probs, dict)
    assert isinstance(dist_entropy, dict)
    assert isinstance(state_values, dict)

    for agent_id in agent_ids:
        assert agent_id in actions
        assert agent_id in log_probs
        assert agent_id in dist_entropy
        assert agent_id in state_values



import random
from numbers import Number
from typing import Any, Optional, Union

import numpy as np
import torch
import torch.nn as nn
from gymnasium import spaces
from tensordict import TensorDict

from agilerl.components.data import Transition
from agilerl.modules import EvolvableModule
from agilerl.typing import NumpyObsType, TorchObsType


def assert_state_dicts_equal(
    state_dict1: dict[str, torch.Tensor],
    state_dict2: dict[str, torch.Tensor],
    rtol: float = 1e-5,
    atol: float = 1e-8,
) -> None:
    """
    Compare two PyTorch state dictionaries using torch.allclose for efficient comparison.

    :param state_dict1: First state dictionary
    :type state_dict1: dict[str, torch.Tensor]
    :param state_dict2: Second state dictionary
    :type state_dict2: dict[str, torch.Tensor]
    :param rtol: Relative tolerance for torch.allclose
    :type rtol: float
    :param atol: Absolute tolerance for torch.allclose
    :type atol: float
    """
    # First check that they have the same keys
    assert set(state_dict1.keys()) == set(
        state_dict2.keys()
    ), f"State dict keys don't match: {set(state_dict1.keys())} vs {set(state_dict2.keys())}"

    # Then check each tensor
    for key in state_dict1:
        tensor1 = state_dict1[key]
        tensor2 = state_dict2[key]

        if isinstance(tensor1, torch.Tensor) and isinstance(tensor2, torch.Tensor):
            if tensor1.device != tensor2.device:
                tensor1 = tensor1.cpu()
                tensor2 = tensor2.cpu()

            assert (
                tensor1.shape == tensor2.shape
            ), f"Tensors for key '{key}' have different shapes: {tensor1.shape} != {tensor2.shape}"
            assert torch.allclose(
                tensor1, tensor2, rtol=rtol, atol=atol
            ), f"Tensors for key '{key}' are not close enough"


def assert_not_equal_state_dict(
    state_dict1: dict[str, torch.Tensor],
    state_dict2: dict[str, torch.Tensor],
    rtol: float = 1e-5,
    atol: float = 1e-8,
) -> None:
    """
    Compare two PyTorch state dictionaries using torch.allclose for efficient comparison.

    :param state_dict1: First state dictionary
    :type state_dict1: dict[str, torch.Tensor]
    :param state_dict2: Second state dictionary
    :type state_dict2: dict[str, torch.Tensor]
    :param rtol: Relative tolerance for torch.allclose
    :type rtol: float
    :param atol: Absolute tolerance for torch.allclose
    :type atol: float
    """
    try:
        assert_state_dicts_equal(state_dict1, state_dict2, rtol, atol)
    except AssertionError:
        return

    raise AssertionError(f"State dicts are equal: {state_dict1} == {state_dict2}")


def check_equal_params_ind(
    before_ind: Union[nn.Module, EvolvableModule],
    mutated_ind: Union[nn.Module, EvolvableModule],
) -> None:
    before_dict = dict(before_ind.named_parameters())
    after_dict = mutated_ind.named_parameters()
    for key, param in after_dict:
        if key in before_dict:
            old_param = before_dict[key]
            old_size = old_param.data.size()
            new_size = param.data.size()
            if old_size == new_size:
                # If the sizes are the same, just copy the parameter
                param.data = old_param.data
            elif "norm" not in key:
                # Create a slicing index to handle tensors with varying sizes
                slice_index = tuple(
                    slice(0, min(o, n)) for o, n in zip(old_size, new_size)
                )
                assert torch.all(
                    torch.eq(param.data[slice_index], old_param.data[slice_index])
                ), f"Parameter {key} not equal after mutation {mutated_ind.last_mutation_attr}:\n{param.data[slice_index]}\n{old_param.data[slice_index]}"


def unpack_network(model: nn.Sequential) -> list[nn.Module]:
    """Unpacks an nn.Sequential type model"""
    layer_list = []
    for layer in model.children():
        if isinstance(layer, nn.Sequential):
            # If it's an nn.Sequential, recursively unpack its children
            layer_list.extend(unpack_network(layer))
        else:
            if isinstance(layer, nn.Flatten):
                pass
            else:
                layer_list.append(layer)

    return layer_list


def check_models_same(model1: nn.Module, model2: nn.Module) -> bool:
    for p1, p2 in zip(model1.parameters(), model2.parameters()):
        if p1.data.ne(p2.data).sum() > 0:
            return False
    return True


def generate_random_box_space(
    shape: tuple[int, ...], low: Optional[Number] = None, high: Optional[Number] = None
) -> spaces.Box:
    return spaces.Box(
        low=random.randint(0, 5) if low is None else low,
        high=random.randint(6, 10) if high is None else high,
        shape=shape,
        dtype=np.float32,
    )


def generate_discrete_space(n: int) -> spaces.Discrete:
    return spaces.Discrete(n)


def generate_multidiscrete_space(n: int, m: int) -> spaces.MultiDiscrete:
    return spaces.MultiDiscrete([n] * m)


def generate_dict_or_tuple_space(
    n_image: int,
    n_vector: int,
    image_shape: tuple[int, ...] = (3, 32, 32),
    vector_shape: tuple[int] = (4,),
    dict_space: Optional[bool] = True,
) -> Union[spaces.Dict, spaces.Tuple]:

    if dict_space is None:
        dict_space = True if random.random() < 0.5 else False

    image_spaces = [
        generate_random_box_space(image_shape, low=0, high=255) for _ in range(n_image)
    ]
    vector_spaces = [
        generate_random_box_space(vector_shape, low=-1, high=1) for _ in range(n_vector)
    ]

    if dict_space:
        image_spaces = {f"image_{i}": space for i, space in enumerate(image_spaces)}
        vector_spaces = {f"vector_{i}": space for i, space in enumerate(vector_spaces)}
        return spaces.Dict(image_spaces | vector_spaces)

    return spaces.Tuple(image_spaces + vector_spaces)


def generate_multi_agent_box_spaces(
    n_agents: int,
    shape: tuple[int, ...],
    low: Optional[Union[Number, list[Number]]] = -1,
    high: Optional[Union[Number, list[Number]]] = 1,
) -> list[spaces.Box]:
    if isinstance(low, list):
        assert len(low) == n_agents
    if isinstance(high, list):
        assert len(high) == n_agents

    spaces = []
    for i in range(n_agents):
        if isinstance(low, list):
            _low = low[i]
        else:
            _low = low
        if isinstance(high, list):
            _high = high[i]
        else:
            _high = high

        spaces.append(generate_random_box_space(shape, _low, _high))

    return spaces


def generate_multi_agent_discrete_spaces(
    n_agents: int, m: int
) -> list[spaces.Discrete]:
    return [generate_discrete_space(m) for _ in range(n_agents)]


def generate_multi_agent_multidiscrete_spaces(
    n_agents: int, m: int
) -> list[spaces.MultiDiscrete]:
    return [generate_multidiscrete_space(m, m) for _ in range(n_agents)]


def gen_multi_agent_dict_or_tuple_spaces(
    n_agents: int,
    n_image: int,
    n_vector: int,
    image_shape: tuple[int, ...] = (3, 16, 16),
    vector_shape: tuple[int] = (4,),
    dict_space: Optional[bool] = False,
) -> list[Union[spaces.Dict, spaces.Tuple]]:
    return [
        generate_dict_or_tuple_space(
            n_image, n_vector, image_shape, vector_shape, dict_space
        )
        for _ in range(n_agents)
    ]


def get_sample_from_space(
    space: spaces.Space,
    batch_size: Optional[int] = None,
    device: Optional[torch.device] = None,
) -> NumpyObsType:
    """
    Generate a sample from the given space.

    :param space: The space to generate a sample from.
    :type space: spaces.Space
    :param batch_size: The batch size.
    :type batch_size: int
    :return: A sample from the space.
    :rtype: NumpyObsType
    """
    if isinstance(space, spaces.Box):
        if batch_size is None:
            return np.random.uniform(low=space.low, high=space.high, size=space.shape)
        else:
            return np.random.uniform(
                low=space.low, high=space.high, size=(batch_size, *space.shape)
            )
    elif isinstance(space, spaces.Discrete):
        if batch_size is None:
            return np.random.randint(space.n, size=(1,))
        else:
            return np.random.randint(space.n, size=(batch_size, 1))
    elif isinstance(space, spaces.MultiDiscrete):
        if batch_size is None:
            return np.random.randint(space.nvec, size=(len(space.nvec),))
        else:
            return np.random.randint(space.nvec, size=(batch_size, len(space.nvec)))
    elif isinstance(space, spaces.Dict):
        return {
            key: get_sample_from_space(value, batch_size)
            for key, value in space.items()
        }
    elif isinstance(space, spaces.Tuple):
        return tuple(get_sample_from_space(value, batch_size) for value in space)
    else:
        raise ValueError(f"Unsupported space type: {type(space)}")


def is_processed_observation(observation: TorchObsType, space: spaces.Space) -> bool:
    if isinstance(space, spaces.Box):
        return (
            isinstance(observation, torch.Tensor)
            and observation.shape[1:] == space.shape
        )
    elif isinstance(space, spaces.Discrete):
        return isinstance(observation, torch.Tensor) and observation.shape[1:] == (1,)
    elif isinstance(space, spaces.MultiDiscrete):
        return isinstance(observation, torch.Tensor) and observation.shape[1:] == (
            len(space.nvec),
        )
    elif isinstance(space, spaces.Dict):
        return isinstance(observation, dict) and all(
            is_processed_observation(observation[key], space[key])
            for key in space.keys()
        )
    elif isinstance(space, spaces.Tuple):
        return isinstance(observation, tuple) and all(
            is_processed_observation(value, space[i])
            for i, value in enumerate(observation)
        )
    else:
        raise ValueError(f"Unsupported space type: {type(space)}")


def get_experiences_batch(
    observation_space: spaces.Space,
    action_space: spaces.Space,
    batch_size: int,
    device: Optional[torch.device] = None,
) -> TensorDict:
    """
    Generate a batch of experiences from the observation and action spaces.

    :param observation_space: The observation space.
    :type observation_space: spaces.Space
    :param action_space: The action space.
    :type action_space: spaces.Space
    :param batch_size: The batch size.
    :type batch_size: int
    :param device: The device to run the experiences on.
    :type device: torch.device
    :return: A batch of experiences.
    :rtype: TensorDict
    """
    device = device if device is not None else "cpu"
    states = get_sample_from_space(observation_space, batch_size)
    actions = get_sample_from_space(action_space, batch_size)
    rewards = torch.randn((batch_size, 1))
    next_states = get_sample_from_space(observation_space, batch_size)
    dones = torch.randint(0, 2, (batch_size, 1))
    return Transition(
        obs=states,
        action=actions,
        reward=rewards,
        next_obs=next_states,
        done=dones,
        batch_size=[batch_size],
        device=device,
    ).to_tensordict()


def assert_close_dict(before: dict[str, Any], after: dict[str, Any]) -> None:
    for key, value in before.items():
        if isinstance(value, dict):
            assert_close_dict(value, after[key])
        elif isinstance(value, torch.Tensor):
            assert torch.allclose(
                value, after[key]
            ), f"Value not close: {value} != {after[key]}"
        else:
            assert value == after[key], f"Value not equal: {value} != {after[key]}"



from typing import Any, Optional

import gymnasium as gym
import numpy as np
import torch
import torch.nn as nn

from agilerl.modules.base import EvolvableModule, MutationType, mutation
from agilerl.typing import ArrayOrTensor
from agilerl.utils.evolvable_networks import create_mlp


class ClipReward(gym.RewardWrapper):
    def __init__(self, env):
        super().__init__(env)

    def reward(self, reward: float) -> float:
        # Clip the reward to the range (-1, 1)
        return np.sign(float(reward))


class BasicNetActorDQN(nn.Module):
    def __init__(self, input_size, hidden_sizes, output_size):
        super().__init__()
        layers = []

        # Add input layer
        layers.append(nn.Linear(input_size, hidden_sizes[0]))
        layers.append(nn.ReLU())  # Activation function

        # Add hidden layers
        for i in range(len(hidden_sizes) - 1):
            layers.append(nn.Linear(hidden_sizes[i], hidden_sizes[i + 1]))
            layers.append(nn.ReLU())  # Activation function

        # Add output layer with a sigmoid activation
        layers.append(nn.Linear(hidden_sizes[-1], output_size))
        # layers.append(nn.Tanh())  # Sigmoid activation

        # Combine all layers into a sequential model
        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)


class BasicNetActor(nn.Module):
    def __init__(self, input_size, hidden_sizes, output_size):
        super().__init__()
        layers = []

        # Add input layer
        layers.append(nn.Linear(input_size, hidden_sizes[0]))
        layers.append(nn.ReLU())  # Activation function

        # Add hidden layers
        for i in range(len(hidden_sizes) - 1):
            layers.append(nn.Linear(hidden_sizes[i], hidden_sizes[i + 1]))
            layers.append(nn.ReLU())  # Activation function

        # Add output layer with a sigmoid activation
        layers.append(nn.Linear(hidden_sizes[-1], output_size))
        layers.append(nn.Tanh())  # Sigmoid activation

        # Combine all layers into a sequential model
        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)


class SoftmaxActor(nn.Module):
    def __init__(self, input_size, hidden_sizes, output_size):
        super().__init__()
        layers = []

        # Add input layer
        layers.append(nn.Linear(input_size, hidden_sizes[0]))
        layers.append(nn.ReLU())  # Activation function

        # Add hidden layers
        for i in range(len(hidden_sizes) - 1):
            layers.append(nn.Linear(hidden_sizes[i], hidden_sizes[i + 1]))
            layers.append(nn.ReLU())  # Activation function

        # Add output layer with a sigmoid activation
        layers.append(nn.Linear(hidden_sizes[-1], output_size))
        layers.append(nn.Softmax())  # Output activation

        # Combine all layers into a sequential model
        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)


class BasicNetCritic(nn.Module):
    def __init__(self, input_size, hidden_sizes, output_size):
        super().__init__()
        layers = []

        # Add input layer
        layers.append(nn.Linear(input_size, hidden_sizes[0]))
        layers.append(nn.ReLU())  # Activation function

        # Add hidden layers
        for i in range(len(hidden_sizes) - 1):
            layers.append(nn.Linear(hidden_sizes[i], hidden_sizes[i + 1]))
            layers.append(nn.ReLU())  # Activation function

        # Add output layer with a sigmoid activation
        layers.append(nn.Linear(hidden_sizes[-1], output_size))
        # layers.append(nn.Sigmoid())  # Sigmoid activation

        # Combine all layers into a sequential model
        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)


class SimpleCNNActor(nn.Module):
    def __init__(self, num_classes):
        super().__init__()

        # Define the convolutional layers
        self.conv1 = nn.Conv2d(
            in_channels=4, out_channels=16, kernel_size=8, stride=4
        )  # W: 160, H: 210
        self.conv2 = nn.Conv2d(
            in_channels=16, out_channels=32, kernel_size=4, stride=2
        )  # W:

        # Define the max-pooling layers
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)

        # Define fully connected layers
        self.fc1 = nn.Linear(2592, 256)  # Assuming input images are 128x128
        self.fc2 = nn.Linear(256, num_classes)

        # Define activation function
        self.relu = nn.ReLU()

        # Define softmax for classification
        self.softmax = nn.Softmax(dim=1)

        self.tanh = nn.Tanh()

    def forward(self, x):
        # Forward pass through convolutional layers
        x = self.relu(self.conv1(x))
        x = self.relu(self.conv2(x))
        # Flatten the output for the fully connected layers
        x = x.view(x.size(0), -1)

        # Forward pass through fully connected layers
        x = self.relu(self.fc1(x))
        x = self.fc2(x)
        x = self.softmax(x)
        return x


class SimpleCNNCritic(nn.Module):
    def __init__(self, num_classes):
        super().__init__()

        # Define the convolutional layers
        self.conv1 = nn.Conv2d(
            in_channels=4, out_channels=16, kernel_size=8, stride=4
        )  # W: 160, H: 210
        self.conv2 = nn.Conv2d(
            in_channels=16, out_channels=32, kernel_size=4, stride=2
        )  # W:

        # Define the max-pooling layers
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)

        # Define fully connected layers
        self.fc1 = nn.Linear(2592, 256)
        self.fc2 = nn.Linear(256, num_classes)

        # Define activation function
        self.relu = nn.ReLU()

        # Define softmax for classification
        self.softmax = nn.Softmax(dim=1)
        self.tanh = nn.Tanh()

    def forward(self, x):
        # Forward pass through convolutional layers
        x = self.relu(self.conv1(x))
        x = self.relu(self.conv2(x))

        # Flatten the output for the fully connected layers
        x = x.view(x.size(0), -1)

        # Forward pass through fully connected layers
        x = self.tanh(self.fc1(x))
        x = self.fc2(x)

        # Apply softmax for classification
        x = self.softmax(x)

        return x


class MultiCNNActor(nn.Module):
    def __init__(self, num_classes):
        super().__init__()
        # Define the convolutional layers
        self.conv1 = nn.Conv3d(
            in_channels=4, out_channels=16, kernel_size=(1, 3, 3), stride=4
        )  # W: 160, H: 210
        self.conv2 = nn.Conv3d(
            in_channels=16, out_channels=32, kernel_size=(1, 3, 3), stride=2
        )

        # Define the max-pooling layers
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)

        # Define fully connected layers
        self.fc1 = nn.Linear(3200, 256)  # Assuming input images are 128x128
        self.fc2 = nn.Linear(256, num_classes)

        # Define activation function
        self.relu = nn.ReLU()

        # Define softmax for classification
        self.softmax = nn.Softmax(dim=1)

        self.tanh = nn.Tanh()

    def forward(self, x):
        # Forward pass through convolutional layers
        x = self.relu(self.conv1(x))
        x = self.relu(self.conv2(x))
        # Flatten the output for the fully connected layers
        x = x.view(x.size(0), -1)

        # Forward pass through fully connected layers
        x = self.relu(self.fc1(x))
        x = self.fc2(x)
        x = self.softmax(x)
        return x


class MultiCNNCritic(nn.Module):
    def __init__(self, num_classes, num_agents):
        super().__init__()

        # Define the convolutional layers
        self.conv1 = nn.Conv3d(
            in_channels=4, out_channels=16, kernel_size=(2, 3, 3), stride=4
        )  # W: 160, H: 210
        self.conv2 = nn.Conv3d(
            in_channels=16, out_channels=32, kernel_size=(1, 3, 3), stride=2
        )  # W:

        # Define the max-pooling layers
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)

        # Define fully connected layers
        self.fc1 = nn.Linear(3202, 256)
        self.fc2 = nn.Linear(256, num_classes)

        # Define activation function
        self.relu = nn.ReLU()

        # Define softmax for classification
        self.softmax = nn.Softmax(dim=1)
        self.tanh = nn.Tanh()

        self.action_tensor = torch.ones(1, num_agents, device="cuda")

    def forward(self, x, xc):
        # Forward pass through convolutional layers
        x = self.relu(self.conv1(x))
        x = self.relu(self.conv2(x))

        # Flatten the output for the fully connected layers
        x = x.view(x.size(0), -1)
        x = torch.cat([x, xc], dim=1)
        # Forward pass through fully connected layers
        x = self.tanh(self.fc1(x))
        x = self.fc2(x)

        # Apply softmax for classification
        x = self.softmax(x)

        return x


class SimpleCritic(EvolvableModule):
    def __init__(
        self,
        num_inputs: int,
        num_outputs: int,
        hidden_size: list[int],
        activation: str = "ReLU",
        output_activation: str = None,
        min_hidden_layers: int = 1,
        max_hidden_layers: int = 3,
        min_mlp_nodes: int = 64,
        max_mlp_nodes: int = 500,
        layer_norm: bool = True,
        output_vanish: bool = True,
        init_layers: bool = True,
        noisy: bool = False,
        noise_std: float = 0.5,
        new_gelu: bool = False,
        device: str = "cpu",
        name: str = "mlp",
    ):
        super().__init__(device)

        assert (
            num_inputs > 0
        ), "'num_inputs' cannot be less than or equal to zero, please enter a valid integer."
        assert (
            num_outputs > 0
        ), "'num_outputs' cannot be less than or equal to zero, please enter a valid integer."
        for num in hidden_size:
            assert (
                num > 0
            ), "'hidden_size' cannot contain zero, please enter a valid integer."
        assert len(hidden_size) != 0, "MLP must contain at least one hidden layer."
        assert (
            min_hidden_layers < max_hidden_layers
        ), "'min_hidden_layers' must be less than 'max_hidden_layers."
        assert (
            min_mlp_nodes < max_mlp_nodes
        ), "'min_mlp_nodes' must be less than 'max_mlp_nodes."

        self.name = name
        self.num_inputs = num_inputs
        self.num_outputs = num_outputs
        self._activation = activation
        self.new_gelu = new_gelu
        self.output_activation = output_activation
        self.min_hidden_layers = min_hidden_layers
        self.max_hidden_layers = max_hidden_layers
        self.min_mlp_nodes = min_mlp_nodes
        self.max_mlp_nodes = max_mlp_nodes
        self.layer_norm = layer_norm
        self.output_vanish = output_vanish
        self.init_layers = init_layers
        self.hidden_size = hidden_size
        self.noisy = noisy
        self.noise_std = noise_std

        self.model = create_mlp(
            input_size=self.num_inputs,
            output_size=self.num_outputs,
            hidden_size=self.hidden_size,
            output_vanish=self.output_vanish,
            output_activation=self.output_activation,
            noisy=self.noisy,
            init_layers=self.init_layers,
            layer_norm=self.layer_norm,
            activation=self.activation,
            noise_std=self.noise_std,
            device=self.device,
            new_gelu=self.new_gelu,
            name=self.name,
        )

    @property
    def net_config(self) -> dict[str, Any]:
        """Returns model configuration in dictionary."""
        net_config = self.init_dict.copy()
        for attr in ["num_inputs", "num_outputs", "device", "name"]:
            if attr in net_config:
                net_config.pop(attr)

        return net_config

    @property
    def activation(self) -> str:
        """Returns activation function."""
        return self._activation

    @activation.setter
    def activation(self, activation: str) -> None:
        """Set activation function."""
        self._activation = activation

    def init_weights_gaussian(
        self, std_coeff: float = 4, output_coeff: float = 4
    ) -> None:
        """Initialise weights of neural network using Gaussian distribution."""
        EvolvableModule.init_weights_gaussian(self.model, std_coeff=std_coeff)

        # Output layer is initialised with std_coeff=2
        output_layer = self.get_output_dense()
        EvolvableModule.init_weights_gaussian(output_layer, std_coeff=output_coeff)

    def forward(
        self, obs: dict[str, ArrayOrTensor], actions: ArrayOrTensor
    ) -> torch.Tensor:
        """Returns output of neural network.

        :param x: Neural network input
        :type x: torch.Tensor()

        :return: Neural network output
        :rtype: torch.Tensor
        """
        if not isinstance(obs, dict):
            obs = {
                k: torch.tensor(v, dtype=torch.float32, device=self.device)
                for k, v in obs.items()
            }

        if not isinstance(actions, torch.Tensor):
            actions = torch.tensor(actions, dtype=torch.float32, device=self.device)

        if len(next(iter(obs.values())).shape) == 1:
            obs = {k: v.unsqueeze(0) for k, v in obs.items()}

        if len(actions.shape) == 1:
            actions = actions.unsqueeze(0)

        x = torch.cat([torch.cat(list(obs.values()), dim=1), actions], dim=1)
        return self.model(x)

    def get_output_dense(self) -> torch.nn.Module:
        """Returns output layer of neural network."""
        return getattr(self.model, f"{self.name}_linear_layer_output")

    def change_activation(self, activation: str, output: bool = False) -> None:
        """Set the activation function for the network.

        :param activation: Activation function to use.
        :type activation: str
        :param output: Flag indicating whether to set the output activation function, defaults to False
        :type output: bool, optional
        """
        if output:
            self.output_activation = activation

        self.activation = activation
        self.recreate_network()

    @mutation(MutationType.LAYER)
    def add_layer(self) -> None:
        """Adds a hidden layer to neural network. Falls back on ``add_node()`` if
        max hidden layers reached."""
        # add layer to hyper params
        if len(self.hidden_size) < self.max_hidden_layers:  # HARD LIMIT
            self.hidden_size += [self.hidden_size[-1]]
        else:
            return self.add_node()

    @mutation(MutationType.LAYER)
    def remove_layer(self) -> None:
        """Removes a hidden layer from neural network. Falls back on ``add_node()`` if
        min hidden layers reached."""
        if len(self.hidden_size) > self.min_hidden_layers:  # HARD LIMIT
            self.hidden_size = self.hidden_size[:-1]
        else:
            return self.add_node()

    @mutation(MutationType.NODE)
    def add_node(
        self, hidden_layer: Optional[int] = None, numb_new_nodes: Optional[int] = None
    ) -> dict[str, int]:
        """Adds nodes to hidden layer of neural network.

        :param hidden_layer: Depth of hidden layer to add nodes to, defaults to None
        :type hidden_layer: int, optional
        :param numb_new_nodes: Number of nodes to add to hidden layer, defaults to None
        :type numb_new_nodes: int, optional
        """
        if hidden_layer is None:
            hidden_layer = np.random.randint(0, len(self.hidden_size), 1)[0]
        else:
            hidden_layer = min(hidden_layer, len(self.hidden_size) - 1)

        if numb_new_nodes is None:
            numb_new_nodes = np.random.choice([16, 32, 64], 1)[0]

        if (
            self.hidden_size[hidden_layer] + numb_new_nodes <= self.max_mlp_nodes
        ):  # HARD LIMIT
            self.hidden_size[hidden_layer] += numb_new_nodes

        return {"hidden_layer": hidden_layer, "numb_new_nodes": numb_new_nodes}

    @mutation(MutationType.NODE)
    def remove_node(
        self, hidden_layer: Optional[int] = None, numb_new_nodes: Optional[int] = None
    ) -> dict[str, int]:
        """Removes nodes from hidden layer of neural network.

        :param hidden_layer: Depth of hidden layer to remove nodes from, defaults to None
        :type hidden_layer: int, optional
        :param numb_new_nodes: Number of nodes to remove from hidden layer, defaults to None
        :type numb_new_nodes: int, optional
        """
        if hidden_layer is None:
            hidden_layer = np.random.randint(0, len(self.hidden_size), 1)[0]
        else:
            hidden_layer = min(hidden_layer, len(self.hidden_size) - 1)

        if numb_new_nodes is None:
            numb_new_nodes = np.random.choice([16, 32, 64], 1)[0]

        # HARD LIMIT
        if self.hidden_size[hidden_layer] - numb_new_nodes > self.min_mlp_nodes:
            self.hidden_size[hidden_layer] -= numb_new_nodes

        return {"hidden_layer": hidden_layer, "numb_new_nodes": numb_new_nodes}

    def recreate_network(self) -> None:
        """Recreates neural networks.

        :param shrink_params: Shrink parameters of neural networks, defaults to False
        :type shrink_params: bool, optional
        """
        model = create_mlp(
            input_size=self.num_inputs,
            output_size=self.num_outputs,
            hidden_size=self.hidden_size,
            output_vanish=self.output_vanish,
            output_activation=self.output_activation,
            noisy=self.noisy,
            init_layers=self.init_layers,
            layer_norm=self.layer_norm,
            activation=self.activation,
            noise_std=self.noise_std,
            new_gelu=self.new_gelu,
            device=self.device,
            name=self.name,
        )

        self.model = EvolvableModule.preserve_parameters(
            old_net=self.model, new_net=model
        )



import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from agilerl.typing import DeviceType


class GumbelSoftmax(nn.Module):
    """Applies gumbel softmax function element-wise"""

    @staticmethod
    def gumbel_softmax(
        logits: torch.Tensor, tau: float = 1.0, eps: float = 1e-20
    ) -> torch.Tensor:
        """Implementation of the gumbel softmax activation function

        :param logits: Tensor containing unnormalized log probabilities for each class.
        :type logits: torch.Tensor
        :param tau: Tau, defaults to 1.0
        :type tau: float, optional
        :param eps: Epsilon, defaults to 1e-20
        :type eps: float, optional
        """
        torch.compiler.cudagraph_mark_step_begin()
        epsilon = torch.rand_like(logits)  # epsilon = U
        gumbel_noise = -torch.log(-torch.log(epsilon + eps) + eps)
        y = logits + gumbel_noise
        return F.softmax(y / tau, dim=-1)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return self.gumbel_softmax(input)


class NoisyLinear(nn.Module):
    """The Noisy Linear Neural Network class.

    :param in_features: Input features size
    :type in_features: int
    :param out_features: Output features size
    :type out_features: int
    :param std_init: Standard deviation, defaults to 0.5
    :type std_init: float, optional
    :param device: Device, defaults to "cpu"
    :type device: str, optional
    """

    weight_epsilon: torch.Tensor
    bias_epsilon: torch.Tensor

    def __init__(
        self,
        in_features: int,
        out_features: int,
        std_init: float = 0.5,
        device: DeviceType = "cpu",
    ):
        super().__init__()

        self.in_features = in_features
        self.out_features = out_features
        self.std_init = std_init
        self.device = device

        self.weight_mu = nn.Parameter(
            torch.empty(out_features, in_features, device=device)
        )
        self.weight_sigma = nn.Parameter(
            torch.empty(out_features, in_features, device=device)
        )
        self.register_buffer(
            "weight_epsilon", torch.empty(out_features, in_features, device=device)
        )

        self.bias_mu = nn.Parameter(torch.empty(out_features, device=device))
        self.bias_sigma = nn.Parameter(torch.empty(out_features, device=device))
        self.register_buffer("bias_epsilon", torch.empty(out_features, device=device))

        self.reset_parameters()
        self.reset_noise()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(in_features={self.in_features}, out_features={self.out_features})"

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Returns output of neural network.

        :param x: Neural network input
        :type x: torch.Tensor
        :return: Neural network output
        :rtype: torch.Tensor
        """
        if self.training:
            weight = self.weight_mu + self.weight_sigma.mul(self.weight_epsilon)
            bias = self.bias_mu + self.bias_sigma.mul(self.bias_epsilon)
        else:
            weight = self.weight_mu
            bias = self.bias_mu

        return F.linear(x, weight, bias)

    def reset_parameters(self) -> None:
        """Resets neural network parameters."""
        mu_range = 1 / math.sqrt(self.in_features)

        self.weight_mu.data.uniform_(-mu_range, mu_range)
        self.weight_sigma.data.fill_(self.std_init / math.sqrt(self.in_features))

        self.bias_mu.data.uniform_(-mu_range, mu_range)
        self.bias_sigma.data.fill_(self.std_init / math.sqrt(self.out_features))

    def reset_noise(self) -> None:
        """Resets neural network noise."""
        epsilon_in = self._scale_noise(self.in_features)
        epsilon_out = self._scale_noise(self.out_features)

        self.weight_epsilon.copy_(epsilon_out.ger(epsilon_in))
        self.bias_epsilon.copy_(epsilon_out)

    def _scale_noise(self, size: int) -> torch.Tensor:
        """Returns noisy tensor.

        :param size: Tensor of same size as noisy output
        :type size: torch.Tensor()
        """
        x = torch.randn(size, device=self.device)
        return x.sign().mul_(x.abs().sqrt_())


class NewGELU(nn.Module):
    """
    Implementation of the GELU activation function currently in Google BERT repo (identical to OpenAI GPT).
    Reference: Gaussian Error Linear Units (GELU) paper: https://arxiv.org/abs/1606.08415
    """

    def forward(self, x):
        return (
            0.5
            * x
            * (
                1.0
                + torch.tanh(
                    math.sqrt(2.0 / math.pi) * (x + 0.044715 * torch.pow(x, 3.0))
                )
            )
        )


class ResidualBlock(nn.Module):
    """Residual block with support for even and odd kernel sizes.

    :param in_channels: Number of input channels.
    :type in_channels: int
    :param kernel_size: Size of the convolutional kernel.
    :type kernel_size: int
    :param scale_factor: Expansion factor for the hidden layer.
    :type scale_factor: int, optional (default=4)
    :param device: Device for computation.
    :type device: str, optional (default="cpu")
    """

    def __init__(
        self,
        in_channels: int,
        kernel_size: int,
        scale_factor: int = 4,
        device: str = "cpu",
    ) -> None:
        super().__init__()

        hidden_channels = in_channels * scale_factor

        self.conv1 = nn.Conv2d(
            in_channels,
            hidden_channels,
            kernel_size=kernel_size,
            padding=0,
            bias=False,
            device=device,
        )
        self.bn1 = nn.BatchNorm2d(hidden_channels, device=device)

        self.conv2 = nn.Conv2d(
            hidden_channels,
            in_channels,
            kernel_size=kernel_size,
            padding=0,
            bias=False,
            device=device,
        )
        self.bn2 = nn.BatchNorm2d(in_channels, device=device)

        # Weight initialization
        nn.init.kaiming_uniform_(self.conv1.weight, nonlinearity="relu")
        nn.init.kaiming_uniform_(self.conv2.weight, nonlinearity="relu")

    def asymmetric_padding(self, x: torch.Tensor, kernel_size: int) -> torch.Tensor:
        """Applies asymmetric padding for even kernel sizes."""
        pad_l = (kernel_size - 1) // 2  # Floor
        pad_r = kernel_size // 2  # Ceiling
        return F.pad(x, (pad_l, pad_r, pad_l, pad_r), mode="replicate")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        res = x

        x = self.asymmetric_padding(
            x, kernel_size=self.conv1.kernel_size[0]
        )  # Apply manual padding
        x = F.relu(self.bn1(self.conv1(x)))

        x = self.asymmetric_padding(
            x, kernel_size=self.conv2.kernel_size[0]
        )  # Apply manual padding
        x = self.bn2(self.conv2(x))

        return F.relu(res + x)  # Apply ReLU after summing the residual


class SimbaResidualBlock(nn.Module):
    """Creates a residual block designed to avoid overfitting in RL by inducing
    a simplicity bias through skip connections.

    Paper: https://arxiv.org/abs/2410.09754

    :param hidden_size: Hidden size of the residual block
    :type hidden_size: int
    :param scale_factor: Scale factor, defaults to 4
    :type scale_factor: float
    :param device: Device, defaults to "cpu"
    :type device
    """

    def __init__(
        self, hidden_size: int, scale_factor: float = 4, device: DeviceType = "cpu"
    ) -> None:
        super().__init__()

        self.hidden_size = hidden_size

        self.layer_norm = nn.LayerNorm(hidden_size, device=device)
        self.linear1 = nn.Linear(hidden_size, hidden_size * scale_factor, device=device)
        self.linear2 = nn.Linear(hidden_size * scale_factor, hidden_size, device=device)

        # initialize weigts using he initialization
        nn.init.kaiming_uniform_(self.linear1.weight)
        nn.init.kaiming_uniform_(self.linear2.weight)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        res = x
        x = self.layer_norm(x)
        x = F.relu(self.linear1(x))
        x = self.linear2(x)
        return res + x



import numpy as np
import torch

from agilerl.utils.sampling_utils import (
    always_terminate,
    get_relevant_kvs,
    map_decoder_kvs,
    pad_sequence,
    process_logits,
    select_batch_idxs,
    update_decoder_kvs,
    update_kvs,
)


# The function returns a tensor with the same shape as x, but with the elements indexed by idxs in the first dimension.
def test_same_shape_with_indexed_elements():
    x = torch.tensor([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    idxs = torch.tensor([0, 2])
    expected_output = torch.tensor([[1, 2, 3], [7, 8, 9]])
    assert torch.all(select_batch_idxs(x, idxs) == expected_output)


# The function pads a sequence with a given value up to a certain length.
def test_pad_sequence_pads_sequence_with_given_value():
    seq = torch.tensor([1, 2, 3])
    to_len = 5
    val = 0
    device = "cpu"
    dim = 0

    result = pad_sequence(seq, to_len, val, device, dim)

    expected = torch.tensor([1, 2, 3, 0, 0])
    assert torch.all(torch.eq(result, expected))


# The function updates the key-value store (kvs) with the updated key-value store (updated_kvs) for a given index (idx) and chosen length (lens_chosen).
def test_update_kvs_with_updated_kvs():
    kvs = np.array(
        [
            [
                [
                    [[[1, 2], [3, 4]], [[5, 6], [7, 8]]],
                    [[[9, 10], [11, 12]], [[13, 14], [15, 16]]],
                ]
            ]
        ]
    )
    updated_kvs = np.array(
        [
            [
                [
                    [[[17, 18], [19, 20]], [[21, 22], [23, 24]]],
                    [[[25, 26], [27, 28]], [[29, 30], [31, 32]]],
                ]
            ]
        ]
    )
    lens_chosen = slice(0, None)
    idx = 0
    expected_result = np.array(
        [
            [[[17, 18], [3, 4]], [[21, 22], [7, 8]]],
            [[[25, 26], [11, 12]], [[29, 30], [15, 16]]],
        ]
    )

    result = update_kvs(kvs, updated_kvs, lens_chosen, idx)

    assert np.all(result == expected_result)


# The function updates the key-value store (kvs) with the updated key-value store (updated_kvs) for a given index (idx) and chosen length (lens_chosen).
def test_update_decoder_kvs_with_updated_kvs():
    kvs = np.array(
        [
            [
                [
                    [[[1, 2], [3, 4]], [[5, 6], [7, 8]]],
                    [[[9, 10], [11, 12]], [[13, 14], [15, 16]]],
                ]
            ]
        ]
    )
    updated_kvs = np.array(
        [
            [
                [
                    [[[17, 18], [19, 20]], [[21, 22], [23, 24]]],
                    [[[25, 26], [27, 28]], [[29, 30], [31, 32]]],
                ]
            ]
        ]
    )
    lens_chosen = slice(0, None)
    idx = 0
    expected_result = np.array(
        [
            [[[17, 18], [3, 4]], [[21, 22], [7, 8]]],
            [[[25, 26], [11, 12]], [[29, 30], [15, 16]]],
        ]
    )

    result = update_decoder_kvs(kvs, updated_kvs, lens_chosen, idx)

    assert np.all(result == expected_result)


# The function returns the expected output when given valid inputs.
def test_get_relevant_kvs():
    kvs = (
        (
            torch.tensor([[[[[[1, 2, 3], [4, 5, 6]]]]]]),
            torch.tensor([[[[[[7, 8, 9], [10, 11, 12]]]]]]),
        ),
        (
            torch.tensor([[[[[[13, 14, 15], [16, 17, 18]]]]]]),
            torch.tensor([[[[[[19, 20, 21], [22, 23, 24]]]]]]),
        ),
    )
    lens_chosen = torch.tensor([0])
    idx = 2
    expected_output = (
        (
            torch.tensor([[[[[[1, 2, 3], [4, 5, 6]]]]]]),
            torch.tensor([[[[[[7, 8, 9], [10, 11, 12]]]]]]),
        ),
        (
            torch.tensor([[[[[[13, 14, 15], [16, 17, 18]]]]]]),
            torch.tensor([[[[[[19, 20, 21], [22, 23, 24]]]]]]),
        ),
    )

    assert str(get_relevant_kvs(kvs, lens_chosen, idx)) == str(expected_output)


def test_map_decoder_kvs():
    kvs = (
        (
            torch.tensor([[[[[[1, 2, 3], [4, 5, 6]]]]]]),
            torch.tensor([[[[[[7, 8, 9], [10, 11, 12]]]]]]),
        ),
        (
            torch.tensor([[[[[[13, 14, 15], [16, 17, 18]]]]]]),
            torch.tensor([[[[[[19, 20, 21], [22, 23, 24]]]]]]),
        ),
    )
    lens_chosen = torch.tensor([0])
    kvs = map_decoder_kvs(lambda x: select_batch_idxs(x, lens_chosen), kvs)

    expected_output = (
        (
            torch.tensor([[[[[[1, 2, 3], [4, 5, 6]]]]]]),
            torch.tensor([[[[[[7, 8, 9], [10, 11, 12]]]]]]),
        ),
        (
            torch.tensor([[[[[[13, 14, 15], [16, 17, 18]]]]]]),
            torch.tensor([[[[[[19, 20, 21], [22, 23, 24]]]]]]),
        ),
    )

    assert str(kvs) == str(expected_output)


def test_process_logits():
    logits = torch.randn(8, 4, 8, 4)

    out_logits = process_logits(logits, top_k=2, top_p=2)

    assert logits.shape == out_logits.shape


def test_always_terminate():
    assert always_terminate(np.array([0]))



import numpy as np
import torch

from agilerl.components.data import ReplayDataset, Transition
from agilerl.components.replay_buffer import ReplayBuffer


# The dataset can be initialized with a buffer and batch size
def test_initialization_with_buffer_and_batch_size():
    buffer = ReplayBuffer(
        max_size=1000,
    )
    batch_size = 2
    dataset = ReplayDataset(buffer, batch_size=batch_size)
    assert dataset.buffer == buffer
    assert dataset.batch_size == batch_size


# Sampling a batch of experiences from the buffer works correctly
def test_sampling_batch_from_buffer():
    buffer = ReplayBuffer(
        max_size=1000,
    )

    state1 = np.array([1, 2, 3])
    action1 = np.array([0])
    reward1 = np.array([1])
    next_state1 = np.array([4, 5, 6])
    done1 = np.array([False])

    transition1 = Transition(
        obs=state1,
        action=action1,
        reward=reward1,
        next_obs=next_state1,
        done=done1,
    ).to_tensordict()

    transition1 = transition1.unsqueeze(0)
    transition1.batch_size = [1]
    buffer.add(transition1)
    buffer.add(transition1)

    batch_size = 2
    dataset = ReplayDataset(buffer, batch_size=batch_size)
    iterator = iter(dataset)
    batch = next(iterator)

    assert len(batch) == batch_size
    assert len(batch["obs"]) == batch_size
    assert torch.equal(batch["obs"][0], torch.from_numpy(state1).float())
    assert torch.equal(batch["obs"][1], torch.from_numpy(state1).float())
    assert torch.equal(batch["action"][0], torch.from_numpy(action1).float())
    assert torch.equal(batch["action"][1], torch.from_numpy(action1).float())
    assert torch.equal(batch["reward"][0], torch.from_numpy(reward1).float())
    assert torch.equal(batch["reward"][1], torch.from_numpy(reward1).float())
    assert torch.equal(batch["next_obs"][0], torch.from_numpy(next_state1).float())
    assert torch.equal(batch["next_obs"][1], torch.from_numpy(next_state1).float())
    assert torch.equal(batch["done"][0], torch.from_numpy(done1).float())
    assert torch.equal(batch["done"][1], torch.from_numpy(done1).float())



import os

import h5py
import numpy as np
from accelerate import Accelerator
from tensordict import TensorDictBase
from torch.utils.data import DataLoader

from agilerl.algorithms import CQN
from agilerl.components.data import ReplayDataset, Transition
from agilerl.components.replay_buffer import ReplayBuffer
from agilerl.components.sampler import Sampler
from agilerl.hpo.mutation import Mutations
from agilerl.hpo.tournament import TournamentSelection
from agilerl.utils.utils import (
    create_population,
    default_progress_bar,
    make_vect_envs,
)

# !Note: If you are running this demo without having installed agilerl,
# uncomment and place the following above agilerl imports:

# import sys
# sys.path.append('../')


if __name__ == "__main__":
    accelerator = Accelerator()

    accelerator.wait_for_everyone()
    if accelerator.is_main_process:
        print("===== AgileRL Offline Distributed Demo =====")
    accelerator.wait_for_everyone()

    NET_CONFIG = {
        "encoder_config": {
            "hidden_size": [32, 32],  # Actor hidden size
        }
    }

    INIT_HP = {
        "DOUBLE": True,  # Use double Q-learning
        "BATCH_SIZE": 128,  # Batch size
        "LR": 1e-3,  # Learning rate
        "GAMMA": 0.99,  # Discount factor
        "LEARN_STEP": 1,  # Learning frequency
        "TAU": 1e-3,  # For soft update of target network parameters
        "POP_SIZE": 4,  # Population size
    }

    num_envs = 1
    env = make_vect_envs("CartPole-v1")  # Create environment
    dataset = h5py.File("data/cartpole/cartpole_random_v1.1.0.h5", "r")  # Load dataset

    observation_space = env.single_observation_space
    action_space = env.single_action_space

    pop: list[CQN] = create_population(
        algo="CQN",  # Algorithm
        observation_space=observation_space,  # Observation space
        action_space=action_space,  # Action space
        net_config=NET_CONFIG,  # Network configuration
        INIT_HP=INIT_HP,  # Initial hyperparameters
        population_size=INIT_HP["POP_SIZE"],  # Population size
        num_envs=num_envs,  # Number of vectorized envs
        accelerator=accelerator,  # Accelerator
    )

    field_names = ["state", "action", "reward", "next_state", "done"]
    memory = ReplayBuffer(
        memory_size=10000,  # Max replay buffer size
        field_names=field_names,
    )  # Field names to store in memory

    if accelerator.is_main_process:
        print("Filling replay buffer with dataset...")
    accelerator.wait_for_everyone()

    # Save transitions to replay buffer
    dataset_length = dataset["rewards"].shape[0]

    for i in range(dataset_length - 1):
        obs = dataset["observations"][i]
        next_obs = dataset["observations"][i + 1]
        action = dataset["actions"][i]
        reward = dataset["rewards"][i]
        done = bool(dataset["terminals"][i])
        # Save experience to replay buffer
        transition: TensorDictBase = Transition(obs, action, reward, next_obs, done)
        transition = transition.to_tensordict()
        transition.batch_size = [1]
        memory.add(transition)

    # Create dataloader from replay buffer
    replay_dataset = ReplayDataset(memory, INIT_HP["BATCH_SIZE"])
    replay_dataloader = DataLoader(replay_dataset, batch_size=None)
    replay_dataloader = accelerator.prepare(replay_dataloader)
    sampler = Sampler(
        distributed=True, dataset=replay_dataset, dataloader=replay_dataloader
    )

    tournament = TournamentSelection(
        tournament_size=2,  # Tournament selection size
        elitism=True,  # Elitism in tournament selection
        population_size=INIT_HP["POP_SIZE"],  # Population size
        eval_loop=1,  # Evaluate using last N fitness scores
    )

    mutations = Mutations(
        algo="CQN",  # Algorithm
        no_mutation=0.4,  # No mutation
        architecture=0.2,  # Architecture mutation
        new_layer_prob=0.2,  # New layer mutation
        parameters=0.2,  # Network parameters mutation
        activation=0,  # Activation layer mutation
        rl_hp=0.2,  # Learning HP mutation
        rl_hp_selection=["lr", "batch_size"],  # Learning HPs to choose from
        mutation_sd=0.1,  # Mutation strength  # Network architecture
        rand_seed=1,  # Random seed
        accelerator=accelerator,  # Accelerator
    )

    max_steps = 50000  # Max steps

    evo_steps = 5000  # Evolution frequency
    eval_steps = None  # Evaluation steps per episode - go until done
    eval_loop = 1  # Number of evaluation episodes

    total_steps = 0

    accel_temp_models_path = "models/{}".format("CartPole-v1")
    if accelerator.is_main_process:
        if not os.path.exists(accel_temp_models_path):
            os.makedirs(accel_temp_models_path)

    print(f"\nDistributed training on {accelerator.device}...")

    # TRAINING LOOP
    print("Training...")
    pbar = default_progress_bar(max_steps, accelerator=accelerator)
    while np.less([agent.steps[-1] for agent in pop], max_steps).all():
        if accelerator is not None:
            accelerator.wait_for_everyone()
        for agent in pop:  # Loop through population
            for idx_step in range(evo_steps):
                # Sample dataloader
                experiences = sampler.sample(agent.batch_size)
                # Learn according to agent's RL algorithm
                agent.learn(experiences)
            total_steps += evo_steps
            agent.steps[-1] += evo_steps
            pbar.update(evo_steps)

        # Evaluate population
        fitnesses = [
            agent.test(
                env,
                max_steps=eval_steps,
                loop=eval_loop,
            )
            for agent in pop
        ]

        if accelerator.is_main_process:
            pbar.write(
                f"--- Global steps {total_steps} ---\n"
                f"Steps: {[agent.steps[-1] for agent in pop]}\n"
                f"Fitnesses: {['%.2f' % fitness for fitness in fitnesses]}\n"
                f"5 fitness avgs: {['%.2f' % np.mean(agent.fitness[-5:]) for agent in pop]}\n"
            )

        # Tournament selection and population mutation
        accelerator.wait_for_everyone()
        for model in pop:
            model.unwrap_models()
        accelerator.wait_for_everyone()
        if accelerator.is_main_process:
            elite, pop = tournament.select(pop)
            pop = mutations.mutation(pop)
            for pop_i, model in enumerate(pop):
                model.save_checkpoint(f"{accel_temp_models_path}/CQN_{pop_i}.pt")
        accelerator.wait_for_everyone()
        if not accelerator.is_main_process:
            for pop_i, model in enumerate(pop):
                model.load_checkpoint(f"{accel_temp_models_path}/CQN_{pop_i}.pt")
        accelerator.wait_for_everyone()
        for model in pop:
            model.wrap_models()

    pbar.close()
    env.close()



from typing import ClassVar

import gymnasium as gym
import numpy as np
import torch
import yaml

from agilerl.algorithms.core.registry import HyperparameterConfig, RLParameter
from agilerl.hpo.mutation import Mutations
from agilerl.hpo.tournament import TournamentSelection
from agilerl.training.train_on_policy import train_on_policy
from agilerl.utils.utils import (
    create_population,
    make_vect_envs,
    print_hyperparams,
)
from agilerl.wrappers.agent import RSNorm

# !Note: If you are running this demo without having installed agilerl,
# uncomment and place the following above agilerl imports:

# import sys
# sys.path.append('../')


class MaskVelocityWrapper(gym.ObservationWrapper):
    """
    Gym environment observation wrapper used to mask velocity terms in
    observations. The intention is the make the MDP partially observable.
    Adapted from https://github.com/LiuWenlin595/FinalProject.

    Taken from https://github.com/DLR-RM/rl-baselines3-zoo/blob/master/rl_zoo3/wrappers.py#L299.

    :param env: Gym environment
    :type env: gym.Env
    """

    # Supported envs
    velocity_indices: ClassVar[dict[str, np.ndarray]] = {
        "CartPole-v1": np.array([1, 3]),
        "MountainCar-v0": np.array([1]),
        "MountainCarContinuous-v0": np.array([1]),
        "Pendulum-v1": np.array([2]),
        "LunarLander-v3": np.array([2, 3, 5]),
        "LunarLanderContinuous-v3": np.array([2, 3, 5]),
    }

    def __init__(self, env: gym.Env):
        super().__init__(env)

        assert env.unwrapped.spec is not None
        env_id: str = env.unwrapped.spec.id
        # By default no masking
        self.mask = np.ones_like(env.observation_space.sample())
        try:
            # Mask velocity
            self.mask[self.velocity_indices[env_id]] = 0.0
        except KeyError as e:
            raise NotImplementedError(
                f"Velocity masking not implemented for {env_id}"
            ) from e

    def observation(self, observation: np.ndarray) -> np.ndarray:
        return observation * self.mask


def main_recurrent(INIT_HP, MUTATION_PARAMS, NET_CONFIG):
    # Set device
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    print("============ AgileRL ============")
    print(f"DEVICE: {device}")

    def make_env():
        return MaskVelocityWrapper(gym.make(INIT_HP["ENV_NAME"]))

    env = make_vect_envs(make_env=make_env, num_envs=INIT_HP["NUM_ENVS"])

    observation_space = env.single_observation_space
    action_space = env.single_action_space

    tournament = TournamentSelection(
        INIT_HP["TOURN_SIZE"],
        INIT_HP["ELITISM"],
        INIT_HP["POP_SIZE"],
        INIT_HP["EVAL_LOOP"],
    )
    mutations = Mutations(
        no_mutation=MUTATION_PARAMS["NO_MUT"],
        architecture=MUTATION_PARAMS["ARCH_MUT"],
        new_layer_prob=MUTATION_PARAMS["NEW_LAYER"],
        parameters=MUTATION_PARAMS["PARAMS_MUT"],
        activation=MUTATION_PARAMS["ACT_MUT"],
        rl_hp=MUTATION_PARAMS["RL_HP_MUT"],
        mutation_sd=MUTATION_PARAMS["MUT_SD"],
        rand_seed=MUTATION_PARAMS["RAND_SEED"],
        device=device,
    )

    hp_config = HyperparameterConfig(
        lr=RLParameter(min=MUTATION_PARAMS["MIN_LR"], max=MUTATION_PARAMS["MAX_LR"]),
        batch_size=RLParameter(
            min=MUTATION_PARAMS["MIN_BATCH_SIZE"],
            max=MUTATION_PARAMS["MAX_BATCH_SIZE"],
            dtype=int,
        ),
        # learn_step=RLParameter(
        #     min=MUTATION_PARAMS["MIN_LEARN_STEP"],
        #     max=MUTATION_PARAMS["MAX_LEARN_STEP"],
        #     dtype=int,
        #     grow_factor=1.5,
        #     shrink_factor=0.75,
        # ),
        ent_coef=RLParameter(
            min=MUTATION_PARAMS["MIN_ENT_COEF"],
            max=MUTATION_PARAMS["MAX_ENT_COEF"],
            grow_factor=1.0,
            shrink_factor=0.9,
        ),
    )

    agent_pop = create_population(
        algo=INIT_HP["ALGO"],
        observation_space=observation_space,
        action_space=action_space,
        net_config=NET_CONFIG,
        INIT_HP=INIT_HP,
        hp_config=hp_config,
        population_size=INIT_HP["POP_SIZE"],
        num_envs=INIT_HP["NUM_ENVS"],
        device=device,
    )

    # Normalize observations using running mean and std
    agent_pop = [RSNorm(agent) for agent in agent_pop]

    env_name_prefix = INIT_HP["ENV_NAME"].split("-")[0]
    env_name_prefix = env_name_prefix + "NoVel"
    env_name = env_name_prefix + "-" + INIT_HP["ENV_NAME"].split("-")[1]
    trained_pop, pop_fitnesses = train_on_policy(
        env,
        env_name,
        INIT_HP["ALGO"],
        agent_pop,
        INIT_HP=INIT_HP,
        MUT_P=MUTATION_PARAMS,
        swap_channels=INIT_HP["CHANNELS_LAST"],
        max_steps=INIT_HP["MAX_STEPS"],
        evo_steps=INIT_HP["EVO_STEPS"],
        eval_steps=INIT_HP["EVAL_STEPS"],
        eval_loop=INIT_HP["EVAL_LOOP"],
        target=INIT_HP["TARGET_SCORE"],
        tournament=tournament,
        mutation=mutations,
        wb=INIT_HP["WANDB"],
    )

    print_hyperparams(trained_pop)
    # plot_population_score(trained_pop)

    if str(device) == "cuda":
        torch.cuda.empty_cache()

    env.close()


if __name__ == "__main__":
    with open("configs/training/ppo/ppo_recurrent.yaml") as file:
        config = yaml.safe_load(file)
    INIT_HP = config["INIT_HP"]
    MUTATION_PARAMS = config["MUTATION_PARAMS"]
    NET_CONFIG = config["NET_CONFIG"]
    main_recurrent(INIT_HP, MUTATION_PARAMS, NET_CONFIG)
